import warnings

import nibabel as nib
import numpy as np
import numpy.testing as npt

from dipy.core.gradients import gradient_table
from dipy.core.sphere import HemiSphere, unit_octahedron
from dipy.data import get_fnames, get_sphere
from dipy.direction import (
    BootDirectionGetter,
    ClosestPeakDirectionGetter,
    DeterministicMaximumDirectionGetter,
    PTTDirectionGetter,
    PeaksAndMetrics,
    ProbabilisticDirectionGetter,
)
from dipy.reconst.csdeconv import ConstrainedSphericalDeconvModel
from dipy.reconst.shm import descoteaux07_legacy_msg
from dipy.sims.voxel import multi_tensor, single_tensor
from dipy.testing.decorators import set_random_number_generator
from dipy.tracking.local_tracking import LocalTracking, ParticleFilteringTracking
from dipy.tracking.stopping_criterion import (
    ActStoppingCriterion,
    BinaryStoppingCriterion,
    StreamlineStatus,
    ThresholdStoppingCriterion,
)
from dipy.tracking.streamline import Streamlines
from dipy.tracking.utils import seeds_from_mask


def allclose(x, y, atol=None):
    if atol is not None:
        return x.shape == y.shape and np.allclose(x, y, atol=0.5)
    else:
        return x.shape == y.shape and np.allclose(x, y)


def test_stop_conditions():
    """This tests that the Local Tracker behaves as expected for the
    following tissue types.
    """
    # StreamlineStatus.TRACKPOINT = 1
    # StreamlineStatus.ENDPOINT = 2
    # StreamlineStatus.INVALIDPOINT = 0
    tissue = np.array(
        [
            [2, 1, 1, 2, 1],
            [2, 2, 1, 1, 2],
            [1, 1, 1, 1, 1],
            [1, 1, 1, 2, 2],
            [0, 1, 1, 1, 2],
            [0, 1, 1, 0, 2],
            [1, 0, 1, 1, 1],
            [2, 1, 2, 0, 0],
        ]
    )
    tissue = tissue[None]

    sphere = HemiSphere.from_sphere(unit_octahedron)
    pmf_lookup = np.array(
        [
            [
                0.0,
                0.0,
                0.0,
            ],
            [0.0, 0.0, 1.0],
        ]
    )
    pmf = pmf_lookup[(tissue > 0).astype("int")]

    # Create a seeds along
    x = np.array([0.0, 0, 0, 0, 0, 0, 0, 0])
    y = np.array([0.0, 1, 2, 3, 4, 5, 6, 7])
    z = np.array([1.0, 1, 1, 0, 1, 1, 1, 1])
    seeds = np.column_stack([x, y, z])

    # Set up tracking
    endpoint_mask = tissue == StreamlineStatus.ENDPOINT
    invalidpoint_mask = tissue == StreamlineStatus.INVALIDPOINT
    sc = ActStoppingCriterion(endpoint_mask, invalidpoint_mask)
    dg = ProbabilisticDirectionGetter.from_pmf(pmf, 60, sphere)

    # valid streamlines only
    streamlines_generator = LocalTracking(
        direction_getter=dg,
        stopping_criterion=sc,
        seeds=seeds,
        affine=np.eye(4),
        step_size=1.0,
        return_all=False,
    )
    streamlines_not_all = iter(streamlines_generator)

    # all streamlines
    streamlines_all_generator = LocalTracking(
        direction_getter=dg,
        stopping_criterion=sc,
        seeds=seeds,
        affine=np.eye(4),
        step_size=1.0,
        return_all=True,
    )
    streamlines_all = iter(streamlines_all_generator)

    # Check that the first streamline stops at 1 and 2 (ENDPOINT)
    y = 0
    sl = next(streamlines_not_all)
    npt.assert_equal(sl[0], [0, y, 1])
    npt.assert_equal(sl[-1], [0, y, 2])
    npt.assert_equal(len(sl), 2)

    sl = next(streamlines_all)
    npt.assert_equal(sl[0], [0, y, 1])
    npt.assert_equal(sl[-1], [0, y, 2])
    npt.assert_equal(len(sl), 2)

    # Check that the next streamline stops at 1 and 3 (ENDPOINT)
    y = 1
    sl = next(streamlines_not_all)
    npt.assert_equal(sl[0], [0, y, 1])
    npt.assert_equal(sl[-1], [0, y, 3])
    npt.assert_equal(len(sl), 3)

    sl = next(streamlines_all)
    npt.assert_equal(sl[0], [0, y, 1])
    npt.assert_equal(sl[-1], [0, y, 3])
    npt.assert_equal(len(sl), 3)

    # This streamline should be the same as above. This row does not have
    # ENDPOINTs, but the streamline should stop at the edge and not include
    # OUTSIDEIMAGE points.
    y = 2
    sl = next(streamlines_not_all)
    npt.assert_equal(sl[0], [0, y, 0])
    npt.assert_equal(sl[-1], [0, y, 4])
    npt.assert_equal(len(sl), 5)

    sl = next(streamlines_all)
    npt.assert_equal(sl[0], [0, y, 0])
    npt.assert_equal(sl[-1], [0, y, 4])
    npt.assert_equal(len(sl), 5)

    # If we seed on the edge, the first (or last) point in the streamline
    # should be the seed.
    y = 3
    sl = next(streamlines_not_all)
    npt.assert_equal(sl[0], seeds[y])

    sl = next(streamlines_all)
    npt.assert_equal(sl[0], seeds[y])

    # The last 3 seeds should not produce streamlines,
    # INVALIDPOINT streamlines are rejected (return_all=False).
    npt.assert_equal(len(list(streamlines_not_all)), 0)

    # The last 3 seeds should produce invalid streamlines,
    # INVALIDPOINT streamlines are kept (return_all=True).
    # The streamline stops at 1 (INVALIDPOINT) and 3 (ENDPOINT)
    y = 4
    sl = next(streamlines_all)
    npt.assert_equal(sl[0], [0, y, 1])
    npt.assert_equal(sl[-1], [0, y, 3])
    npt.assert_equal(len(sl), 3)

    # The streamline stops at 0 (INVALIDPOINT) and 2 (INVALIDPOINT)
    y = 5
    sl = next(streamlines_all)
    npt.assert_equal(sl[0], [0, y, 1])
    npt.assert_equal(sl[-1], [0, y, 2])
    npt.assert_equal(len(sl), 2)

    # The streamline should contain only one point, the seed point,
    # because no valid initial direction was returned.
    y = 6
    sl = next(streamlines_all)
    npt.assert_equal(sl[0], seeds[y])
    npt.assert_equal(sl[-1], seeds[y])
    npt.assert_equal(len(sl), 1)

    # The streamline should contain only one point, the seed point,
    # because no valid neighboring voxel (ENDPOINT)
    y = 7
    sl = next(streamlines_all)
    npt.assert_equal(sl[0], seeds[y])
    npt.assert_equal(sl[-1], seeds[y])
    npt.assert_equal(len(sl), 1)


def test_save_seeds():
    tissue = np.array(
        [
            [2, 1, 1, 2, 1],
            [2, 2, 1, 1, 2],
            [1, 1, 1, 1, 1],
            [1, 1, 1, 2, 2],
            [0, 1, 1, 1, 2],
            [0, 1, 1, 0, 2],
            [1, 0, 1, 1, 1],
        ]
    )
    tissue = tissue[None]

    sphere = HemiSphere.from_sphere(unit_octahedron)
    pmf_lookup = np.array(
        [
            [
                0.0,
                0.0,
                0.0,
            ],
            [0.0, 0.0, 1.0],
        ]
    )
    pmf = pmf_lookup[(tissue > 0).astype("int")]

    # Create a seeds along
    x = np.array([0.0, 0, 0, 0, 0, 0, 0])
    y = np.array([0.0, 1, 2, 3, 4, 5, 6])
    z = np.array([1.0, 1, 1, 0, 1, 1, 1])
    seeds = np.column_stack([x, y, z])

    # Set up tracking
    endpoint_mask = tissue == StreamlineStatus.ENDPOINT
    invalidpoint_mask = tissue == StreamlineStatus.INVALIDPOINT
    sc = ActStoppingCriterion(endpoint_mask, invalidpoint_mask)
    dg = ProbabilisticDirectionGetter.from_pmf(pmf, 60, sphere)

    # valid streamlines only
    streamlines_generator = LocalTracking(
        direction_getter=dg,
        stopping_criterion=sc,
        seeds=seeds,
        affine=np.eye(4),
        step_size=1.0,
        return_all=False,
        save_seeds=True,
    )

    streamlines_not_all = iter(streamlines_generator)
    # Verify that seeds are returned by the LocalTracker
    _, seed = next(streamlines_not_all)
    npt.assert_equal(seed, seeds[0])
    _, seed = next(streamlines_not_all)
    npt.assert_equal(seed, seeds[1])
    # Verify that seeds are returned by the PFTTracker also
    pft_streamlines = ParticleFilteringTracking(
        direction_getter=dg,
        stopping_criterion=sc,
        seeds=seeds,
        affine=np.eye(4),
        step_size=1.0,
        max_cross=1,
        return_all=False,
        save_seeds=True,
    )
    streamlines = iter(pft_streamlines)
    _, seed = next(streamlines)
    npt.assert_equal(seed, seeds[0])
    _, seed = next(streamlines)
    npt.assert_equal(seed, seeds[1])


@set_random_number_generator(0)
def test_tracking_max_angle(rng):
    """This tests that the angle between streamline points is always smaller
    then the input `max_angle` parameter.
    """

    def get_min_cos_similarity(streamlines):
        min_cos_sim = 1
        for sl in streamlines:
            if len(sl) > 1:
                v = sl[:-1] - sl[1:]  # vectors have norm of 1
                for i in range(len(v) - 1):
                    cos_sim = np.dot(v[i], v[i + 1])
                    if cos_sim < min_cos_sim:
                        min_cos_sim = cos_sim
        return min_cos_sim

    for sphere in [
        get_sphere(name="repulsion100"),
        HemiSphere.from_sphere(get_sphere(name="repulsion100")),
    ]:
        shape_img = [5, 5, 5]
        shape_img.extend([sphere.vertices.shape[0]])
        mask = np.ones(shape_img[:3])
        affine = np.eye(4)
        random_pmf = rng.random(shape_img)
        seeds = seeds_from_mask(mask, affine, density=1)
        sc = ActStoppingCriterion.from_pve(
            mask, np.zeros(shape_img[:3]), np.zeros(shape_img[:3])
        )
        max_angle = 20
        step_size = 1
        dg = ProbabilisticDirectionGetter.from_pmf(
            random_pmf, max_angle, sphere, pmf_threshold=0.1
        )
        # local tracking
        streamlines = Streamlines(LocalTracking(dg, sc, seeds, affine, step_size))
        min_cos_sim = get_min_cos_similarity(streamlines)
        npt.assert_(np.arccos(min_cos_sim) <= np.deg2rad(max_angle))

        # PFT tracking
        streamlines = Streamlines(ParticleFilteringTracking(dg, sc, seeds, affine, 1.0))
        min_cos_sim = get_min_cos_similarity(streamlines)
        npt.assert_(np.arccos(min_cos_sim) <= np.deg2rad(max_angle))


def test_probabilistic_odf_weighted_tracker():
    """This tests that the Probabilistic Direction Getter plays nice
    LocalTracking and produces reasonable streamlines in a simple example.
    """
    sphere = HemiSphere.from_sphere(unit_octahedron)

    # A simple image with three possible configurations, a vertical tract,
    # a horizontal tract and a crossing
    pmf_lookup = np.array(
        [[0.0, 0.0, 1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.6, 0.4, 0.0]]
    )
    simple_image = np.array(
        [
            [0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
            [0, 3, 2, 2, 2, 0],
            [0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
        ]
    )

    simple_image = simple_image[..., None]
    pmf = pmf_lookup[simple_image]

    seeds = [np.array([1.0, 1.0, 0.0])] * 30

    mask = (simple_image > 0).astype(float)
    sc = ThresholdStoppingCriterion(mask, 0.5)

    dg = ProbabilisticDirectionGetter.from_pmf(pmf, 90, sphere, pmf_threshold=0.1)
    streamlines = LocalTracking(dg, sc, seeds, np.eye(4), 1.0)

    expected = [
        np.array(
            [
                [0.0, 1.0, 0.0],
                [1.0, 1.0, 0.0],
                [2.0, 1.0, 0.0],
                [2.0, 2.0, 0.0],
                [2.0, 3.0, 0.0],
                [2.0, 4.0, 0.0],
            ]
        ),
        np.array(
            [
                [0.0, 1.0, 0.0],
                [1.0, 1.0, 0.0],
                [2.0, 1.0, 0.0],
                [3.0, 1.0, 0.0],
                [4.0, 1.0, 0.0],
            ]
        ),
    ]

    path = [False, False]
    for sl in streamlines:
        if allclose(sl, expected[0]):
            path[0] = True
        elif allclose(sl, expected[1]):
            path[1] = True
        else:
            raise AssertionError()
    npt.assert_(all(path))

    # The first path is not possible if 90 degree turns are excluded
    dg = ProbabilisticDirectionGetter.from_pmf(pmf, 80, sphere, pmf_threshold=0.1)
    streamlines = LocalTracking(dg, sc, seeds, np.eye(4), 1.0)

    for sl in streamlines:
        npt.assert_(np.allclose(sl, expected[1]))

    # The first path is not possible if pmf_threshold > 0.67
    # 0.4/0.6 < 2/3, multiplying the pmf should not change the ratio
    dg = ProbabilisticDirectionGetter.from_pmf(10 * pmf, 90, sphere, pmf_threshold=0.67)
    streamlines = LocalTracking(dg, sc, seeds, np.eye(4), 1.0)

    for sl in streamlines:
        npt.assert_(np.allclose(sl, expected[1]))

    # Test non WM seed position
    seeds = [[0, 0, 0], [5, 5, 5]]
    streamlines = LocalTracking(
        dg, sc, seeds, np.eye(4), 0.2, max_cross=1, return_all=True
    )
    streamlines = Streamlines(streamlines)
    npt.assert_(len(streamlines[0]) == 1)  # INVALIDPOINT
    npt.assert_(len(streamlines[1]) == 1)  # OUTSIDEIMAGE

    # Test that all points are within the image volume
    seeds = seeds_from_mask(np.ones(mask.shape), np.eye(4), density=2)
    streamline_generator = LocalTracking(dg, sc, seeds, np.eye(4), 0.5, return_all=True)
    streamlines = Streamlines(streamline_generator)
    for s in streamlines:
        npt.assert_(np.all((s + 0.5).astype(int) >= 0))
        npt.assert_(np.all((s + 0.5).astype(int) < mask.shape))
    # Test that the number of streamline return with return_all=True equal the
    # number of seeds places

    npt.assert_(np.array([len(streamlines) == len(seeds)]))

    # Test reproducibility
    tracking_1 = Streamlines(
        LocalTracking(dg, sc, seeds, np.eye(4), 0.5, random_seed=0)
    )._data
    tracking_2 = Streamlines(
        LocalTracking(dg, sc, seeds, np.eye(4), 0.5, random_seed=0)
    )._data
    npt.assert_equal(tracking_1, tracking_2)


@set_random_number_generator(0)
def test_particle_filtering_tractography(rng):
    """This tests that the ParticleFilteringTracking produces
    more streamlines connecting the gray matter than LocalTracking.
    """
    sphere = get_sphere(name="repulsion100")
    step_size = 0.2

    # Simple tissue masks
    simple_wm = np.array(
        [
            [0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0],
            [0, 1, 1, 1, 0, 0],
            [0, 1, 1, 1, 0, 0],
            [0, 0, 0, 0, 0, 0],
        ]
    )
    simple_wm = np.dstack(
        [
            np.zeros(simple_wm.shape),
            simple_wm,
            simple_wm,
            simple_wm,
            np.zeros(simple_wm.shape),
        ]
    )
    simple_gm = np.array(
        [
            [1, 1, 0, 0, 0, 0],
            [1, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 1, 0],
            [0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0],
        ]
    )
    simple_gm = np.dstack(
        [
            np.zeros(simple_gm.shape),
            simple_gm,
            simple_gm,
            simple_gm,
            np.zeros(simple_gm.shape),
        ]
    )
    simple_csf = np.ones(simple_wm.shape) - simple_wm - simple_gm

    sc = ActStoppingCriterion.from_pve(simple_wm, simple_gm, simple_csf)
    seeds = seeds_from_mask(simple_wm, np.eye(4), density=2)

    # Random pmf in every voxel
    shape_img = list(simple_wm.shape)
    shape_img.extend([sphere.vertices.shape[0]])
    pmf = rng.random(shape_img)

    # Test that PFT recover equal or more streamlines than localTracking
    dg = ProbabilisticDirectionGetter.from_pmf(pmf, 60, sphere)
    local_streamlines_generator = LocalTracking(
        dg, sc, seeds, np.eye(4), step_size, max_cross=1, return_all=False
    )
    local_streamlines = Streamlines(local_streamlines_generator)

    pft_streamlines_generator = ParticleFilteringTracking(
        dg,
        sc,
        seeds,
        np.eye(4),
        step_size,
        max_cross=1,
        return_all=False,
        pft_back_tracking_dist=1,
        pft_front_tracking_dist=0.5,
    )
    pft_streamlines = Streamlines(pft_streamlines_generator)

    npt.assert_(np.array([len(pft_streamlines) > 0]))
    npt.assert_(np.array([len(pft_streamlines) >= len(local_streamlines)]))

    # Test PFT with a PTT direction getter
    dg_ptt = PTTDirectionGetter.from_pmf(pmf, 60, sphere)
    pft_ptt_streamlines = Streamlines(
        ParticleFilteringTracking(
            dg_ptt,
            sc,
            seeds,
            np.eye(4),
            step_size,
            max_cross=1,
            return_all=False,
            pft_back_tracking_dist=1,
            pft_front_tracking_dist=0.5,
        )
    )

    npt.assert_(np.array([len(pft_ptt_streamlines) > 0]))
    npt.assert_(np.array([len(pft_ptt_streamlines) >= len(local_streamlines)]))

    # Test that all points are equally spaced
    for ell in [2, 3, 5, 10, 100]:
        pft_streamlines = ParticleFilteringTracking(
            dg,
            sc,
            seeds,
            np.eye(4),
            step_size,
            max_cross=1,
            return_all=True,
            maxlen=ell,
        )
        for s in pft_streamlines:
            for i in range(len(s) - 1):
                npt.assert_almost_equal(np.linalg.norm(s[i] - s[i + 1]), step_size)

    # Test that all points are within the image volume
    seeds = seeds_from_mask(np.ones(simple_wm.shape), np.eye(4), density=1)
    pft_streamlines_generator = ParticleFilteringTracking(
        dg, sc, seeds, np.eye(4), step_size, max_cross=1, return_all=True
    )
    pft_streamlines = Streamlines(pft_streamlines_generator)

    for s in pft_streamlines:
        npt.assert_(np.all((s + 0.5).astype(int) >= 0))
        npt.assert_(np.all((s + 0.5).astype(int) < simple_wm.shape))

    # Test that the number of streamline return with return_all=True equal the
    # number of seeds places
    npt.assert_(np.array([len(pft_streamlines) == len(seeds)]))

    # Test min and max length
    pft_streamlines_generator = ParticleFilteringTracking(
        dg, sc, seeds, np.eye(4), step_size, maxlen=20, minlen=3, return_all=False
    )
    pft_streamlines = Streamlines(pft_streamlines_generator)

    for s in pft_streamlines:
        npt.assert_(len(s) >= 3)
        npt.assert_(len(s) <= 20)

    # Test non WM seed position
    seeds = [[0, 5, 4], [0, 0, 1], [50, 50, 50]]
    pft_streamlines_generator = ParticleFilteringTracking(
        dg, sc, seeds, np.eye(4), step_size, max_cross=1, return_all=True
    )
    pft_streamlines = Streamlines(pft_streamlines_generator)

    npt.assert_equal(len(pft_streamlines[0]), 3)  # INVALIDPOINT
    npt.assert_equal(len(pft_streamlines[1]), 3)  # ENDPOINT
    npt.assert_equal(len(pft_streamlines[2]), 1)  # OUTSIDEIMAGE

    # Test with wrong StoppingCriterion type
    sc_bin = BinaryStoppingCriterion(simple_wm)
    npt.assert_raises(
        ValueError,
        lambda: ParticleFilteringTracking(dg, sc_bin, seeds, np.eye(4), step_size),
    )
    # Test with invalid back/front tracking distances
    npt.assert_raises(
        ValueError,
        lambda: ParticleFilteringTracking(
            dg,
            sc,
            seeds,
            np.eye(4),
            step_size,
            pft_back_tracking_dist=0,
            pft_front_tracking_dist=0,
        ),
    )
    npt.assert_raises(
        ValueError,
        lambda: ParticleFilteringTracking(
            dg, sc, seeds, np.eye(4), step_size, pft_back_tracking_dist=-1
        ),
    )
    npt.assert_raises(
        ValueError,
        lambda: ParticleFilteringTracking(
            dg,
            sc,
            seeds,
            np.eye(4),
            step_size,
            pft_back_tracking_dist=0,
            pft_front_tracking_dist=-2,
        ),
    )

    # Test with invalid affine shape
    npt.assert_raises(
        ValueError,
        lambda: ParticleFilteringTracking(dg, sc, seeds, np.eye(3), step_size),
    )

    # Test with invalid maxlen
    npt.assert_raises(
        ValueError,
        lambda: ParticleFilteringTracking(
            dg, sc, seeds, np.eye(4), step_size, maxlen=0
        ),
    )
    npt.assert_raises(
        ValueError,
        lambda: ParticleFilteringTracking(
            dg, sc, seeds, np.eye(4), step_size, maxlen=-1
        ),
    )

    # Test with invalid particle count
    npt.assert_raises(
        ValueError,
        lambda: ParticleFilteringTracking(
            dg, sc, seeds, np.eye(4), step_size, particle_count=0
        ),
    )
    npt.assert_raises(
        ValueError,
        lambda: ParticleFilteringTracking(
            dg, sc, seeds, np.eye(4), step_size, particle_count=-1
        ),
    )

    # Test reproducibility
    tracking1 = Streamlines(
        ParticleFilteringTracking(dg, sc, seeds, np.eye(4), step_size, random_seed=0)
    )._data
    tracking2 = Streamlines(
        ParticleFilteringTracking(dg, sc, seeds, np.eye(4), step_size, random_seed=0)
    )._data
    npt.assert_equal(tracking1, tracking2)

    # Test min_wm_pve_before_stopping parameter
    expected = [
        np.array([[1.0, 0.0, 1.0], [1.0, 1.0, 1.0], [1.0, 2.0, 1.0]]),
        np.array(
            [
                [1.0, 0.0, 1.0],
                [1.0, 1.0, 1.0],
                [1.0, 2.0, 1.0],
                [1.0, 3.0, 1.0],
                [1.0, 4.0, 1.0],
            ]
        ),
    ]

    simple_wm = np.array(
        [[0, 0, 0, 0, 0, 0], [0, 0.4, 0.4, 1, 0.4, 0], [0, 0, 0, 0, 0, 0]]
    )
    simple_wm = np.dstack(
        [
            np.zeros(simple_wm.shape),
            simple_wm,
            simple_wm,
            simple_wm,
            np.zeros(simple_wm.shape),
        ]
    )
    simple_gm = np.array(
        [[0, 0, 0, 0, 0, 0], [1, 0.6, 0.6, 0, 0.6, 1], [0, 0, 0, 0, 0, 0]]
    )
    simple_gm = np.dstack(
        [
            np.zeros(simple_gm.shape),
            simple_gm,
            simple_gm,
            simple_gm,
            np.zeros(simple_gm.shape),
        ]
    )
    simple_csf = np.ones(simple_wm.shape) - simple_wm - simple_gm
    sc = ActStoppingCriterion.from_pve(simple_wm, simple_gm, simple_csf)
    seeds = np.array([[1, 1, 1]])
    sphere = HemiSphere.from_sphere(unit_octahedron)
    pmf = np.zeros(list(simple_gm.shape) + [3])
    pmf[:, :, :, 1] = 1  # horizontal bundle

    dg = ProbabilisticDirectionGetter.from_pmf(pmf, 30, sphere)

    pft_streamlines_generator = ParticleFilteringTracking(
        dg,
        sc,
        seeds,
        np.eye(4),
        step_size=1,
        max_cross=1,
        return_all=True,
        min_wm_pve_before_stopping=0,
    )
    pft_streamlines = Streamlines(pft_streamlines_generator)
    npt.assert_(np.allclose(pft_streamlines[0], expected[0]))

    pft_streamlines_generator = ParticleFilteringTracking(
        dg,
        sc,
        seeds,
        np.eye(4),
        step_size=1,
        max_cross=1,
        return_all=True,
        min_wm_pve_before_stopping=1,
    )
    pft_streamlines = Streamlines(pft_streamlines_generator)
    npt.assert_(np.allclose(pft_streamlines[0], expected[1]))

    # Test invalid min_wm_pve_before_stopping parameters
    npt.assert_raises(
        ValueError,
        lambda: ParticleFilteringTracking(
            dg, sc, seeds, np.eye(4), step_size, min_wm_pve_before_stopping=-1
        ),
    )
    npt.assert_raises(
        ValueError,
        lambda: ParticleFilteringTracking(
            dg, sc, seeds, np.eye(4), step_size, min_wm_pve_before_stopping=2
        ),
    )


def test_maximum_deterministic_tracker():
    """This tests that the Maximum Deterministic Direction Getter plays nice
    LocalTracking and produces reasonable streamlines in a simple example.
    """
    sphere = HemiSphere.from_sphere(unit_octahedron)

    # A simple image with three possible configurations, a vertical tract,
    # a horizontal tract and a crossing
    pmf_lookup = np.array(
        [[0.0, 0.0, 1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.4, 0.6, 0.0]]
    )
    simple_image = np.array(
        [
            [0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
            [0, 3, 2, 2, 2, 0],
            [0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
        ]
    )

    simple_image = simple_image[..., None]
    pmf = pmf_lookup[simple_image]

    seeds = [np.array([1.0, 1.0, 0.0])] * 30

    mask = (simple_image > 0).astype(float)
    sc = ThresholdStoppingCriterion(mask, 0.5)

    dg = DeterministicMaximumDirectionGetter.from_pmf(
        pmf, max_angle=100, sphere=sphere, pmf_threshold=0.1
    )
    streamlines = LocalTracking(dg, sc, seeds, np.eye(4), 1.0)

    expected = [
        np.array(
            [
                [0.0, 1.0, 0.0],
                [1.0, 1.0, 0.0],
                [2.0, 1.0, 0.0],
                [2.0, 2.0, 0.0],
                [2.0, 3.0, 0.0],
                [2.0, 4.0, 0.0],
            ]
        ),
        np.array(
            [
                [0.0, 1.0, 0.0],
                [1.0, 1.0, 0.0],
                [2.0, 1.0, 0.0],
                [3.0, 1.0, 0.0],
                [4.0, 1.0, 0.0],
            ]
        ),
        np.array([[0.0, 1.0, 0.0], [1.0, 1.0, 0.0], [2.0, 1.0, 0.0]]),
    ]

    for sl in streamlines:
        if not allclose(sl, expected[0]):
            raise AssertionError()

    # The first path is not possible if 90 degree turns are excluded
    dg = DeterministicMaximumDirectionGetter.from_pmf(
        pmf, 80, sphere, pmf_threshold=0.1
    )
    streamlines = LocalTracking(dg, sc, seeds, np.eye(4), 1.0)

    for sl in streamlines:
        npt.assert_(np.allclose(sl, expected[1]))

    # Both path are not possible if 90 degree turns are exclude and
    # if pmf_threshold is larger than 0.67. Streamlines should stop at
    # the crossing.
    # 0.4/0.6 < 2/3, multiplying the pmf should not change the ratio
    dg = DeterministicMaximumDirectionGetter.from_pmf(
        10 * pmf, 80, sphere, pmf_threshold=0.67
    )
    streamlines = LocalTracking(dg, sc, seeds, np.eye(4), 1.0)

    for sl in streamlines:
        npt.assert_(np.allclose(sl, expected[2]))


def test_bootstap_peak_tracker():
    """This tests that the Bootstrap Peak Direction Getter plays nice
    LocalTracking and produces reasonable streamlines in a simple example.
    """
    sphere = get_sphere(name="repulsion100")

    # A simple image with three possible configurations, a vertical tract,
    # a horizontal tract and a crossing
    simple_image = np.array(
        [
            [0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
            [2, 3, 2, 2, 2, 0],
            [0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
        ]
    )
    simple_image = simple_image[..., None]

    bvecs = sphere.vertices
    bvals = np.ones(len(bvecs)) * 1000
    bvecs = np.insert(bvecs, 0, np.array([0, 0, 0]), axis=0)
    bvals = np.insert(bvals, 0, 0)
    gtab = gradient_table(bvals, bvecs=bvecs)
    angles = [(90, 90), (90, 0)]
    fracs = [50, 50]
    mevals = np.array([[1.5, 0.4, 0.4], [1.5, 0.4, 0.4]]) * 1e-3
    mevecs = [
        np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
        np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]]),
    ]
    voxel1 = single_tensor(gtab, 1, evals=mevals[0], evecs=mevecs[0], snr=None)
    voxel2 = single_tensor(gtab, 1, evals=mevals[0], evecs=mevecs[1], snr=None)
    voxel3, _ = multi_tensor(gtab, mevals, fractions=fracs, angles=angles, snr=None)
    data = np.tile(voxel3, [5, 6, 1, 1])
    data[simple_image == 1] = voxel1
    data[simple_image == 2] = voxel2

    response = (np.array(mevals[1]), 1)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        csd_model = ConstrainedSphericalDeconvModel(gtab, response, sh_order_max=6)

    seeds = [np.array([0.0, 1.0, 0.0]), np.array([2.0, 4.0, 0.0])]

    sc = BinaryStoppingCriterion((simple_image > 0).astype(float))
    sphere = HemiSphere.from_sphere(get_sphere(name="symmetric724"))
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        boot_dg = BootDirectionGetter.from_data(data, csd_model, 60, sphere=sphere)

    streamlines_generator = LocalTracking(boot_dg, sc, seeds, np.eye(4), 1.0)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        streamlines = Streamlines(streamlines_generator)
    expected = [
        np.array(
            [
                [0.0, 1.0, 0.0],
                [1.0, 1.0, 0.0],
                [2.0, 1.0, 0.0],
                [3.0, 1.0, 0.0],
                [4.0, 1.0, 0.0],
            ]
        ),
        np.array(
            [
                [2.0, 4.0, 0.0],
                [2.0, 3.0, 0.0],
                [2.0, 2.0, 0.0],
                [2.0, 1.0, 0.0],
                [2.0, 0.0, 0.0],
            ]
        ),
    ]

    if not allclose(streamlines[0], expected[0], atol=0.5):
        raise AssertionError()
    if not allclose(streamlines[1], expected[1], atol=0.5):
        raise AssertionError()


def test_closest_peak_tracker():
    """This tests that the Closest Peak Direction Getter plays nice
    LocalTracking and produces reasonable streamlines in a simple example.
    """
    sphere = HemiSphere.from_sphere(unit_octahedron)

    # A simple image with three possible configurations, a vertical tract,
    # a horizontal tract and a crossing
    pmf_lookup = np.array(
        [[0.0, 0.0, 1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.5, 0.5, 0.0]]
    )
    simple_image = np.array(
        [
            [0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
            [2, 3, 2, 2, 2, 0],
            [0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
        ]
    )

    simple_image = simple_image[..., None]
    pmf = pmf_lookup[simple_image]

    seeds = [np.array([1.0, 1.0, 0.0]), np.array([2.0, 4.0, 0.0])]

    mask = (simple_image > 0).astype(float)
    sc = BinaryStoppingCriterion(mask)

    dg = ClosestPeakDirectionGetter.from_pmf(pmf, 90, sphere, pmf_threshold=0.1)
    streamlines = Streamlines(LocalTracking(dg, sc, seeds, np.eye(4), 1.0))

    expected = [
        np.array(
            [
                [0.0, 1.0, 0.0],
                [1.0, 1.0, 0.0],
                [2.0, 1.0, 0.0],
                [3.0, 1.0, 0.0],
                [4.0, 1.0, 0.0],
            ]
        ),
        np.array(
            [
                [2.0, 0.0, 0.0],
                [2.0, 1.0, 0.0],
                [2.0, 2.0, 0.0],
                [2.0, 3.0, 0.0],
                [2.0, 4.0, 0.0],
            ]
        ),
    ]

    if not allclose(streamlines[0], expected[0]):
        raise AssertionError()
    if not allclose(streamlines[1], expected[1]):
        raise AssertionError()


def test_eudx_tracker():
    """This tests that the Peaks And Metrics Direction Getter plays nice
    LocalTracking and produces reasonable streamlines in a simple example.
    """
    sphere = HemiSphere.from_sphere(unit_octahedron)

    # A simple image with three possible configurations, a vertical tract,
    # a horizontal tract and a crossing
    peaks_values_lookup = np.array([[0.0, 0.0], [1.0, 0.0], [1.0, 0.0], [0.5, 0.5]])
    peaks_indices_lookup = np.array([[-1, -1], [0, -1], [1, -1], [0, 1]])
    # EuDXDirectionGetter needs at 3 slices on each axis to work
    simple_image = np.zeros([5, 6, 3], dtype=int)
    simple_image[:, :, 1] = np.array(
        [
            [0, 1, 0, 1, 0, 0],
            [0, 1, 0, 1, 0, 0],
            [0, 3, 2, 2, 2, 0],
            [0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
        ]
    )

    dg = PeaksAndMetrics()
    dg.sphere = sphere
    dg.peak_values = peaks_values_lookup[simple_image]
    dg.peak_indices = peaks_indices_lookup[simple_image]
    dg.ang_thr = 90

    mask = (simple_image >= 0).astype(float)
    sc = ThresholdStoppingCriterion(mask, 0.5)
    seeds = [
        np.array([1.0, 1.0, 1.0]),
        np.array([2.0, 4.0, 1.0]),
        np.array([1.0, 3.0, 1.0]),
        np.array([4.0, 4.0, 1.0]),
    ]

    streamlines = LocalTracking(dg, sc, seeds, np.eye(4), 1.0)

    expected = [
        np.array(
            [
                [0.0, 1.0, 1.0],
                [1.0, 1.0, 1.0],
                [2.0, 1.0, 1.0],
                [3.0, 1.0, 1.0],
                [4.0, 1.0, 1.0],
            ]
        ),
        np.array(
            [
                [2.0, 0.0, 1.0],
                [2.0, 1.0, 1.0],
                [2.0, 2.0, 1.0],
                [2.0, 3.0, 1.0],
                [2.0, 4.0, 1.0],
                [2.0, 5.0, 1.0],
            ]
        ),
        np.array(
            [
                [0.0, 3.0, 1.0],
                [1.0, 3.0, 1.0],
                [2.0, 3.0, 1.0],
                [2.0, 4.0, 1.0],
                [2.0, 5.0, 1.0],
            ]
        ),
        np.array([[4.0, 4.0, 1.0]]),
    ]

    for i, sl in enumerate(streamlines):
        npt.assert_(np.allclose(sl, expected[i]))


def test_affine_transformations():
    """This tests that the input affine is properly handled by
    LocalTracking and produces reasonable streamlines in a simple example.
    """
    sphere = HemiSphere.from_sphere(unit_octahedron)

    # A simple image with three possible configurations, a vertical tract,
    # a horizontal tract and a crossing
    pmf_lookup = np.array(
        [[0.0, 0.0, 1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.4, 0.6, 0.0]]
    )
    simple_image = np.array(
        [
            [0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
            [0, 3, 2, 2, 2, 0],
            [0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0],
        ]
    )

    simple_image = simple_image[..., None]
    pmf = pmf_lookup[simple_image]

    seeds = [np.array([1.0, 1.0, 0.0]), np.array([2.0, 4.0, 0.0])]

    expected = [
        np.array([[1.0, 1.0, 0.0], [2.0, 1.0, 0.0], [3.0, 1.0, 0.0]]),
        np.array([[2.0, 1.0, 0.0], [2.0, 2.0, 0.0], [2.0, 3.0, 0.0], [2.0, 4.0, 0.0]]),
    ]

    mask = (simple_image > 0).astype(float)
    sc = BinaryStoppingCriterion(mask)

    dg = DeterministicMaximumDirectionGetter.from_pmf(
        pmf, 60, sphere, pmf_threshold=0.1
    )

    # TST- bad affine wrong shape
    bad_affine = np.eye(3)
    npt.assert_raises(ValueError, LocalTracking, dg, sc, seeds, bad_affine, 1.0)

    # TST - bad affine with shearing
    bad_affine = np.eye(4)
    bad_affine[0, 1] = 1.0
    npt.assert_raises(ValueError, LocalTracking, dg, sc, seeds, bad_affine, 1.0)

    # TST - bad seeds
    bad_seeds = 1000
    npt.assert_raises(ValueError, LocalTracking, dg, sc, bad_seeds, np.eye(4), 1.0)

    # TST - identity
    a0 = np.eye(4)
    # TST - affines with positive/negative offsets
    a1 = np.eye(4)
    a1[:3, 3] = [1, 2, 3]
    a2 = np.eye(4)
    a2[:3, 3] = [-2, 0, -1]
    # TST - affine with scaling
    a3 = np.eye(4)
    a3[0, 0] = a3[1, 1] = a3[2, 2] = 8
    # TST - affine with axes inverting (negative value)
    a4 = np.eye(4)
    a4[1, 1] = a4[2, 2] = -1
    # TST - combined affines
    a5 = a1 + a2 + a3
    a5[3, 3] = 1
    # TST - in vivo affine example
    # Sometimes data have affines with tiny shear components.
    # For example, the small_101D data-set has some of that:
    fdata, _, _ = get_fnames(name="small_101D")
    a6 = nib.load(fdata).affine

    for affine in [a0, a1, a2, a3, a4, a5, a6]:
        lin = affine[:3, :3]
        offset = affine[:3, 3]
        seeds_trans = [np.dot(lin, s) + offset for s in seeds]

        # We compute the voxel size to adjust the step size to one voxel
        voxel_size = np.mean(np.sqrt(np.dot(lin, lin).diagonal()))

        streamlines = LocalTracking(
            direction_getter=dg,
            stopping_criterion=sc,
            seeds=seeds_trans,
            affine=affine,
            step_size=voxel_size,
            return_all=True,
        )

        # We apply the inverse affine transformation to the generated
        # streamlines. It should be equals to the expected streamlines
        # (generated with the identity affine matrix).
        affine_inv = np.linalg.inv(affine)
        lin = affine_inv[:3, :3]
        offset = affine_inv[:3, 3]
        streamlines_inv = []
        for line in streamlines:
            streamlines_inv.append([np.dot(pts, lin) + offset for pts in line])

        npt.assert_equal(len(streamlines_inv[0]), len(expected[0]))
        npt.assert_(np.allclose(streamlines_inv[0], expected[0], atol=0.3))
        npt.assert_equal(len(streamlines_inv[1]), len(expected[1]))
        npt.assert_(np.allclose(streamlines_inv[1], expected[1], atol=0.3))


@set_random_number_generator()
def test_random_seed_initialization(rng):
    """Test that the random generator can be initialized correctly with the
    tracking seeds.
    """
    sphere = HemiSphere.from_sphere(unit_octahedron)
    pmf = np.zeros((4, 4, 4, 3))
    x = np.array([0.0, 0, 0, 57.421434502602544])
    y = np.array([0.0, 1, 2, 21.566539227085478])
    z = np.array([1.0, 1, 1, 51.67881720942744])

    seeds = np.vstack([np.column_stack([x, y, z]), rng.random((10, 3))])
    sc = BinaryStoppingCriterion(np.ones((4, 4, 4)))
    dg = ProbabilisticDirectionGetter.from_pmf(pmf, 60, sphere)

    randoms_seeds = (
        [None, 0, 1, -1, np.iinfo(np.uint32).max + 1]
        + list(rng.random(10))
        + list(rng.integers(0, np.iinfo(np.int32).max, 10))
    )

    for rdm_seed in randoms_seeds:
        _ = Streamlines(
            LocalTracking(
                direction_getter=dg,
                stopping_criterion=sc,
                seeds=seeds,
                affine=np.eye(4),
                step_size=1.0,
                random_seed=rdm_seed,
            )
        )


def test_tracking_with_initial_directions():
    """This tests that tractography play well with using seeding directions."""

    def allclose(x, y):
        return x.shape == y.shape and np.allclose(x, y)

    sphere = HemiSphere.from_sphere(unit_octahedron)
    # A simple image with three possible configurations, a vertical tract,
    # a horizontal tract and a crossing
    pmf_lookup = np.array(
        [[0.0, 0.0, 1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.6, 0.4, 0.0]]
    )
    simple_image = np.array(
        [
            [0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
            [2, 3, 2, 2, 2, 2],
            [0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
        ]
    )
    simple_image = simple_image[..., None]
    simple_wm = np.array(
        [
            [0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
            [1, 1, 1, 1, 1, 1],
            [0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
        ]
    )
    simple_wm = simple_wm[..., None]
    simple_csf = np.ones(simple_wm.shape) - simple_wm
    simple_gm = np.zeros(simple_wm.shape)
    pmf = pmf_lookup[simple_image]
    mask = (simple_image > 0).astype(float)
    expected = [
        np.array(
            [
                [0.0, 1.0, 0.0],
                [1.0, 1.0, 0.0],
                [2.0, 1.0, 0.0],
                [2.0, 2.0, 0.0],
                [2.0, 3.0, 0.0],
                [2.0, 4.0, 0.0],
                [2.0, 5.0, 0.0],
            ]
        ),
        np.array(
            [
                [0.0, 1.0, 0.0],
                [1.0, 1.0, 0.0],
                [2.0, 1.0, 0.0],
                [3.0, 1.0, 0.0],
                [4.0, 1.0, 0.0],
            ]
        ),
        np.array(
            [
                [2.0, 0.0, 0.0],
                [2.0, 1.0, 0.0],
                [2.0, 2.0, 0.0],
                [2.0, 3.0, 0.0],
                [2.0, 4.0, 0.0],
                [2.0, 5.0, 0.0],
            ]
        ),
    ]

    # Test LocalTracking with initial directions
    sc = ThresholdStoppingCriterion(mask, 0.5)
    dg = ProbabilisticDirectionGetter.from_pmf(pmf, 45, sphere, pmf_threshold=0.1)
    crossing_pos = np.array([2, 1, 0])
    seeds = np.array([crossing_pos, crossing_pos, crossing_pos])
    initial_directions = np.array(
        [
            [sphere.vertices[0], [0, 0, 0]],
            [sphere.vertices[1], sphere.vertices[0]],
            [[0, 0, 0], [0, 0, 0]],
        ]
    )
    # with max_cross=1
    streamline_generator = LocalTracking(
        dg,
        sc,
        seeds,
        np.eye(4),
        1,
        max_cross=1,
        return_all=True,
        initial_directions=initial_directions,
    )
    streamlines = Streamlines(streamline_generator)
    npt.assert_(allclose(streamlines[0], expected[1]))
    npt.assert_(allclose(streamlines[1], expected[2]))
    npt.assert_(allclose(streamlines[2], np.array([crossing_pos])))
    # with max_cross=2
    streamline_generator = LocalTracking(
        dg,
        sc,
        seeds,
        np.eye(4),
        1,
        max_cross=2,
        return_all=True,
        initial_directions=initial_directions,
    )
    streamlines = Streamlines(streamline_generator)
    npt.assert_(allclose(streamlines[0], expected[1]))
    npt.assert_(allclose(streamlines[1], expected[2]))
    npt.assert_(allclose(streamlines[2], expected[1]))
    npt.assert_(allclose(streamlines[3], np.array([crossing_pos])))

    # Test initial_directions with norm != 1 and not sphere vertices
    initial_directions = np.array(
        [
            [[0, 0, 0], [2, 0, 0]],
            [[0.1, 0.8, 0], [-0.4, 0, 0]],
            [[0, 0, 0], [0.7, 0.6, -0.1]],
        ]
    )
    streamline_generator = LocalTracking(
        dg,
        sc,
        seeds,
        np.eye(4),
        1,
        max_cross=2,
        return_all=False,
        initial_directions=initial_directions,
    )
    streamlines = Streamlines(streamline_generator)
    npt.assert_(allclose(streamlines[0], expected[1]))
    npt.assert_(allclose(streamlines[1], expected[2]))
    npt.assert_(allclose(streamlines[2], expected[1][::-1]))
    npt.assert_(allclose(streamlines[3], expected[1]))

    # Test dimension mismatch between seeds and initial_directions
    npt.assert_raises(
        ValueError,
        lambda: LocalTracking(
            dg,
            sc,
            seeds,
            np.eye(4),
            0.2,
            max_cross=1,
            return_all=True,
            initial_directions=initial_directions[:1, :, :],
        ),
    )

    # Test warning is raised for possible directional biases
    npt.assert_warns(
        Warning,
        lambda: LocalTracking(
            dg,
            sc,
            seeds,
            np.eye(4),
            1,
            max_cross=2,
            return_all=False,
            unidirectional=True,
            randomize_forward_direction=False,
            initial_directions=None,
        ),
    )

    # Test ParticleFilteringTracking with initial directions
    dg = ProbabilisticDirectionGetter.from_pmf(pmf, 45, sphere, pmf_threshold=0.1)
    sc = ActStoppingCriterion.from_pve(simple_wm, simple_gm, simple_csf)
    crossing_pos = np.array([2, 1, 0])
    seeds = np.array([crossing_pos, crossing_pos, crossing_pos])
    initial_directions = np.array(
        [
            [sphere.vertices[0], [0, 0, 0]],
            [sphere.vertices[1], sphere.vertices[0]],
            [[0, 0, 0], [0, 0, 0]],
        ]
    )

    streamline_generator = ParticleFilteringTracking(
        dg,
        sc,
        seeds,
        np.eye(4),
        1,
        return_all=True,
        initial_directions=initial_directions,
    )
    streamlines = Streamlines(streamline_generator)
    npt.assert_(allclose(streamlines[0], expected[1]))
    npt.assert_(allclose(streamlines[1], expected[2]))
    npt.assert_(allclose(streamlines[2], expected[1]))
    npt.assert_(allclose(streamlines[3], np.array([crossing_pos])))

    # Test unidirectional tracking with initial directions
    initial_directions = np.array(
        [[[1, 0, 0], [0, 0, 0]], [[-1, 0, 0], [0, 0, 0]], [[0, -1, 0], [0, 1, 0]]]
    )
    streamline_generator = LocalTracking(
        dg,
        sc,
        seeds,
        np.eye(4),
        1,
        max_cross=2,
        return_all=False,
        unidirectional=True,
        initial_directions=initial_directions,
    )
    streamlines = Streamlines(streamline_generator)
    npt.assert_(allclose(streamlines[0], expected[1][2:]))
    npt.assert_(allclose(streamlines[1], expected[1][:3][::-1]))
    npt.assert_(allclose(streamlines[2], expected[2][:2][::-1]))
    npt.assert_(allclose(streamlines[3], expected[2][1:]))

    streamline_generator = ParticleFilteringTracking(
        dg,
        sc,
        seeds,
        np.eye(4),
        1,
        return_all=True,
        unidirectional=True,
        initial_directions=initial_directions,
    )
    streamlines = Streamlines(streamline_generator)
    npt.assert_(allclose(streamlines[0], expected[1][2:]))
    npt.assert_(allclose(streamlines[1], expected[1][:3][::-1]))
    npt.assert_(allclose(streamlines[2], expected[2][:2][::-1]))
    npt.assert_(allclose(streamlines[3], expected[2][1:]))

    # Test randomized initial forward direction
    seeds = np.array([crossing_pos] * 30)
    initial_directions = np.array([np.array([1, 0, 0])] * 30)[:, np.newaxis, :]
    streamline_generator = LocalTracking(
        dg,
        sc,
        seeds,
        np.eye(4),
        1,
        max_cross=2,
        return_all=False,
        unidirectional=True,
        randomize_forward_direction=True,
        initial_directions=initial_directions,
    )
    streamlines = Streamlines(streamline_generator)
    for sl in streamlines:
        npt.assert_(
            np.allclose(sl, expected[1][2:]) or np.allclose(sl, expected[1][:3][::-1])
        )



"""Benchmarks for ``dipy.segment`` module."""

import numpy as np

from dipy.data import get_fnames
from dipy.io.streamline import load_tractogram
from dipy.segment.clustering import QuickBundles as QB_New
from dipy.segment.mask import bounding_box
from dipy.segment.metricspeed import Metric
from dipy.tracking.streamline import Streamlines, set_number_of_points


class BenchMask:
    def setup(self):
        self.dense_vol = np.zeros((100, 100, 100))
        self.dense_vol[:] = 10
        self.sparse_vol = np.zeros((100, 100, 100))
        self.sparse_vol[0, 0, 0] = 1

    def time_bounding_box_sparse(self):
        bounding_box(self.sparse_vol)

    def time_bounding_box_dense(self):
        bounding_box(self.dense_vol)


class BenchQuickbundles:
    def setup(self):
        dtype = "float32"
        nb_points = 12
        # The expected number of clusters of the fornix using threshold=10
        # is 4.
        self.basic_parameters = {
            "threshold": 10,
            "expected_nb_clusters": 4 * 8,
        }

        fname = get_fnames(name="fornix")

        fornix = load_tractogram(fname, "same", bbox_valid_check=False).streamlines

        fornix_streamlines = Streamlines(fornix)
        fornix_streamlines = set_number_of_points(
            fornix_streamlines, nb_points=nb_points
        )

        # Create eight copies of the fornix to be clustered (one in
        # each octant).
        self.streamlines = []
        self.streamlines += [
            s + np.array([100, 100, 100], dtype) for s in fornix_streamlines
        ]
        self.streamlines += [
            s + np.array([100, -100, 100], dtype) for s in fornix_streamlines
        ]
        self.streamlines += [
            s + np.array([100, 100, -100], dtype) for s in fornix_streamlines
        ]
        self.streamlines += [
            s + np.array([100, -100, -100], dtype) for s in fornix_streamlines
        ]
        self.streamlines += [
            s + np.array([-100, 100, 100], dtype) for s in fornix_streamlines
        ]
        self.streamlines += [
            s + np.array([-100, -100, 100], dtype) for s in fornix_streamlines
        ]
        self.streamlines += [
            s + np.array([-100, 100, -100], dtype) for s in fornix_streamlines
        ]
        self.streamlines += [
            s + np.array([-100, -100, -100], dtype) for s in fornix_streamlines
        ]

        class MDFpy(Metric):
            def are_compatible(self, shape1, shape2):
                return shape1 == shape2

            def dist(self, features1, features2):
                dist = np.sqrt(np.sum((features1 - features2) ** 2, axis=1))
                dist = np.sum(dist / len(features1))
                return dist

        self.custom_metric = MDFpy()

    def time_quickbundles(self):
        qb2 = QB_New(self.basic_parameters.get("threshold", 10))
        _ = qb2.cluster(self.streamlines)

    def time_quickbundles_metric(self):
        qb = QB_New(
            self.basic_parameters.get("threshold", 10), metric=self.custom_metric
        )
        _ = qb.cluster(self.streamlines)



import numpy as np
from numpy.testing import (
    assert_array_almost_equal,
    assert_array_equal,
    assert_equal,
    assert_raises,
)

from dipy.align.transforms import Transform, regtransforms
from dipy.testing.decorators import set_random_number_generator


def test_number_of_parameters():
    expected_params = {
        ("TRANSLATION", 2): 2,
        ("TRANSLATION", 3): 3,
        ("ROTATION", 2): 1,
        ("ROTATION", 3): 3,
        ("RIGID", 2): 3,
        ("RIGID", 3): 6,
        ("SCALING", 2): 1,
        ("SCALING", 3): 1,
        ("AFFINE", 2): 6,
        ("AFFINE", 3): 12,
        ("RIGIDSCALING", 2): 5,
        ("RIGIDSCALING", 3): 9,
        ("RIGIDISOSCALING", 2): 4,
        ("RIGIDISOSCALING", 3): 7,
    }

    for ttype, transform in regtransforms.items():
        assert_equal(transform.get_number_of_parameters(), expected_params[ttype])


@set_random_number_generator()
def test_param_to_matrix_2d(rng):
    # Test translation matrix 2D
    transform = regtransforms[("TRANSLATION", 2)]
    dx, dy = rng.uniform(size=(2,))
    theta = np.array([dx, dy])
    expected = np.array([[1, 0, dx], [0, 1, dy], [0, 0, 1]])
    actual = transform.param_to_matrix(theta)
    assert_array_equal(actual, expected)

    # Test rotation matrix 2D
    transform = regtransforms[("ROTATION", 2)]
    angle = rng.uniform()
    theta = np.array([angle])
    ct = np.cos(angle)
    st = np.sin(angle)
    expected = np.array([[ct, -st, 0], [st, ct, 0], [0, 0, 1]])
    actual = transform.param_to_matrix(theta)
    assert_array_almost_equal(actual, expected)

    # Test rigid matrix 2D
    transform = regtransforms[("RIGID", 2)]
    angle, dx, dy = rng.uniform(size=(3,))
    theta = np.array([angle, dx, dy])
    ct = np.cos(angle)
    st = np.sin(angle)
    expected = np.array([[ct, -st, dx], [st, ct, dy], [0, 0, 1]])
    actual = transform.param_to_matrix(theta)
    assert_array_almost_equal(actual, expected)

    # Test scaling matrix 2D
    transform = regtransforms[("SCALING", 2)]
    factor = rng.uniform()
    theta = np.array([factor])
    expected = np.array([[factor, 0, 0], [0, factor, 0], [0, 0, 1]])
    actual = transform.param_to_matrix(theta)
    assert_array_almost_equal(actual, expected)

    # Test rigid isoscaling matrix 2D
    transform = regtransforms[("RIGIDISOSCALING", 2)]
    angle, dx, dy, factor = rng.uniform(size=(4,))
    theta = np.array([angle, dx, dy, factor])
    ct = np.cos(angle)
    st = np.sin(angle)
    expected = np.array(
        [[ct * factor, -st * factor, dx], [st * factor, ct * factor, dy], [0, 0, 1]]
    )
    actual = transform.param_to_matrix(theta)
    assert_array_almost_equal(actual, expected)

    # Test rigid scaling matrix 2D
    transform = regtransforms[("RIGIDSCALING", 2)]
    angle, dx, dy, sx, sy = rng.uniform(size=(5,))
    theta = np.array([angle, dx, dy, sx, sy])
    ct = np.cos(angle)
    st = np.sin(angle)
    expected = np.array([[ct * sx, -st * sx, dx], [st * sy, ct * sy, dy], [0, 0, 1]])
    actual = transform.param_to_matrix(theta)
    assert_array_almost_equal(actual, expected)

    # Test affine 2D
    transform = regtransforms[("AFFINE", 2)]
    theta = rng.uniform(size=(6,))
    expected = np.eye(3)
    expected[0, :] = theta[:3]
    expected[1, :] = theta[3:6]
    actual = transform.param_to_matrix(theta)
    assert_array_almost_equal(actual, expected)

    # Verify that ValueError is raised if incorrect number of parameters
    for transform in regtransforms.values():
        n = transform.get_number_of_parameters()
        # Set incorrect number of parameters
        theta = np.zeros(n + 1, dtype=np.float64)
        assert_raises(ValueError, transform.param_to_matrix, theta)


@set_random_number_generator()
def test_param_to_matrix_3d(rng):
    # Test translation matrix 3D
    transform = regtransforms[("TRANSLATION", 3)]
    dx, dy, dz = rng.uniform(size=(3,))
    theta = np.array([dx, dy, dz])
    expected = np.array([[1, 0, 0, dx], [0, 1, 0, dy], [0, 0, 1, dz], [0, 0, 0, 1]])
    actual = transform.param_to_matrix(theta)
    assert_array_equal(actual, expected)

    # Create helpful rotation matrix function
    def get_rotation_matrix(alpha, beta, gamma):
        ca = np.cos(alpha)
        sa = np.sin(alpha)
        cb = np.cos(beta)
        sb = np.sin(beta)
        cc = np.cos(gamma)
        sc = np.sin(gamma)

        X = np.array([[1, 0, 0], [0, ca, -sa], [0, sa, ca]])
        Y = np.array([[cb, 0, sb], [0, 1, 0], [-sb, 0, cb]])
        Z = np.array([[cc, -sc, 0], [sc, cc, 0], [0, 0, 1]])

        return Z.dot(X.dot(Y))  # Apply in order: Y, X, Z (Y goes to the right)

    # Test rotation matrix 3D
    transform = regtransforms[("ROTATION", 3)]
    theta = rng.uniform(size=(3,))
    R = get_rotation_matrix(theta[0], theta[1], theta[2])
    expected = np.eye(4)
    expected[:3, :3] = R[:3, :3]
    actual = transform.param_to_matrix(theta)
    assert_array_almost_equal(actual, expected)

    # Test rigid matrix 3D
    transform = regtransforms[("RIGID", 3)]
    theta = rng.uniform(size=(6,))
    R = get_rotation_matrix(theta[0], theta[1], theta[2])
    expected = np.eye(4)
    expected[:3, :3] = R[:3, :3]
    expected[:3, 3] = theta[3:6]
    actual = transform.param_to_matrix(theta)
    assert_array_almost_equal(actual, expected)

    # Test rigid isoscaling matrix 3D
    transform = regtransforms[("RIGIDISOSCALING", 3)]
    theta = rng.uniform(size=(7,))
    R = get_rotation_matrix(theta[0], theta[1], theta[2])
    expected = np.eye(4)
    expected[:3, :3] = R[:3, :3] * theta[6]
    expected[:3, 3] = theta[3:6]
    actual = transform.param_to_matrix(theta)
    assert_array_almost_equal(actual, expected)

    # Test rigid scaling matrix 3D
    transform = regtransforms[("RIGIDSCALING", 3)]
    theta = rng.uniform(size=(9,))
    R = get_rotation_matrix(theta[0], theta[1], theta[2])
    expected = np.eye(4)
    R[0, :3] *= theta[6]
    R[1, :3] *= theta[7]
    R[2, :3] *= theta[8]
    expected[:3, :3] = R[:3, :3]
    expected[:3, 3] = theta[3:6]
    actual = transform.param_to_matrix(theta)
    assert_array_almost_equal(actual, expected)

    # Test scaling matrix 3D
    transform = regtransforms[("SCALING", 3)]
    factor = rng.uniform()
    theta = np.array([factor])
    expected = np.array(
        [[factor, 0, 0, 0], [0, factor, 0, 0], [0, 0, factor, 0], [0, 0, 0, 1]]
    )
    actual = transform.param_to_matrix(theta)
    assert_array_almost_equal(actual, expected)

    # Test affine 3D
    transform = regtransforms[("AFFINE", 3)]
    theta = rng.uniform(size=(12,))
    expected = np.eye(4)
    expected[0, :] = theta[:4]
    expected[1, :] = theta[4:8]
    expected[2, :] = theta[8:12]
    actual = transform.param_to_matrix(theta)
    assert_array_almost_equal(actual, expected)

    # Verify that ValueError is raised if incorrect number of parameters
    for transform in regtransforms.values():
        n = transform.get_number_of_parameters()
        # Set incorrect number of parameters
        theta = np.zeros(n + 1, dtype=np.float64)
        assert_raises(ValueError, transform.param_to_matrix, theta)


def test_identity_parameters():
    for transform in regtransforms.values():
        dim = transform.get_dim()
        theta = transform.get_identity_parameters()

        expected = np.eye(dim + 1)
        actual = transform.param_to_matrix(theta)
        assert_array_almost_equal(actual, expected)


@set_random_number_generator()
def test_jacobian_functions(rng):
    # Compare the analytical Jacobians with their numerical approximations
    h = 1e-8
    nsamples = 50

    for transform in regtransforms.values():
        n = transform.get_number_of_parameters()
        dim = transform.get_dim()

        expected = np.empty((dim, n))
        theta = rng.uniform(size=(n,))
        T = transform.param_to_matrix(theta)

        for _ in range(nsamples):
            x = 255 * (rng.uniform(size=(dim,)) - 0.5)
            actual = transform.jacobian(theta, x)

            # Approximate with finite differences
            x_hom = np.ones(dim + 1)
            x_hom[:dim] = x[:]
            for i in range(n):
                dtheta = theta.copy()
                dtheta[i] += h
                dT = np.array(transform.param_to_matrix(dtheta))
                g = (dT - T).dot(x_hom) / h
                expected[:, i] = g[:dim]

            assert_array_almost_equal(actual, expected, decimal=5)

    # Test ValueError is raised when theta parameter doesn't have the right
    # length
    for transform in regtransforms.values():
        n = transform.get_number_of_parameters()
        # Wrong number of parameters
        theta = np.zeros(n + 1)
        x = np.zeros(dim)
        assert_raises(ValueError, transform.jacobian, theta, x)


def test_invalid_transform():
    # Note: users should not attempt to use the base class Transform:
    # they should get an instance of one of its derived classes from the
    # regtransforms dictionary (the base class is not contained there)
    # If for some reason the user instantiates it and attempts to use it,
    # however, it will raise exceptions when attempting to retrieve its
    # Jacobian, identity parameters or its matrix representation. It will
    # return -1 if queried about its dimension or number of parameters
    transform = Transform()
    theta = np.ndarray(3)
    x = np.ndarray(3)
    assert_raises(ValueError, transform.jacobian, theta, x)
    assert_raises(ValueError, transform.get_identity_parameters)
    assert_raises(ValueError, transform.param_to_matrix, theta)

    expected = -1
    actual = transform.get_number_of_parameters()
    assert_equal(actual, expected)

    actual = transform.get_dim()
    assert_equal(actual, expected)



import tempfile
from urllib.error import HTTPError, URLError

import nibabel as nib
import numpy as np
from numpy.testing import assert_, assert_allclose, assert_array_equal
import pytest
import trx.trx_file_memmap as tmm

from dipy.data import get_fnames
from dipy.io.streamline import load_tractogram
from dipy.io.utils import (
    create_nifti_header,
    decfa,
    decfa_to_float,
    get_reference_info,
    is_reference_info_valid,
    read_img_arr_or_path,
)
from dipy.testing.decorators import set_random_number_generator

FILEPATH_DIX = None


def setup_module():
    global FILEPATH_DIX
    try:
        FILEPATH_DIX, _, _ = get_fnames(name="gold_standard_tracks")
    except (HTTPError, URLError) as e:
        FILEPATH_DIX = None
        error_msg = f'"Tests Data failed to download." Reason: {e}'
        pytest.skip(error_msg, allow_module_level=True)
        return


def teardown_module():
    global FILEPATH_DIX
    FILEPATH_DIX = (None,)


def test_decfa():
    data_orig = np.zeros((4, 4, 4, 3))
    data_orig[0, 0, 0] = np.array([1, 0, 0])
    img_orig = nib.Nifti1Image(data_orig, np.eye(4))
    img_new = decfa(img_orig)
    data_new = np.asanyarray(img_new.dataobj)
    assert data_new[0, 0, 0] == np.array(
        (1, 0, 0), dtype=np.dtype([("R", "uint8"), ("G", "uint8"), ("B", "uint8")])
    )
    assert data_new.dtype == np.dtype([("R", "uint8"), ("G", "uint8"), ("B", "uint8")])

    round_trip = decfa_to_float(img_new)
    data_rt = np.asanyarray(round_trip.dataobj)
    assert np.all(data_rt == data_orig)

    data_orig = np.zeros((4, 4, 4, 3))
    data_orig[0, 0, 0] = np.array([0.1, 0, 0])
    img_orig = nib.Nifti1Image(data_orig, np.eye(4))
    img_new = decfa(img_orig, scale=True)
    data_new = np.asanyarray(img_new.dataobj)
    assert data_new[0, 0, 0] == np.array(
        (25, 0, 0), dtype=np.dtype([("R", "uint8"), ("G", "uint8"), ("B", "uint8")])
    )
    assert data_new.dtype == np.dtype([("R", "uint8"), ("G", "uint8"), ("B", "uint8")])

    round_trip = decfa_to_float(img_new)
    data_rt = np.asanyarray(round_trip.dataobj)
    assert data_rt.shape == (4, 4, 4, 3)
    assert np.all(data_rt[0, 0, 0] == np.array([25, 0, 0]))


def is_affine_valid(affine):
    return is_reference_info_valid(affine, [1, 1, 1], [1.0, 1.0, 1.0], "RAS")


def is_dimensions_valid(dimensions):
    return is_reference_info_valid(np.eye(4), dimensions, [1.0, 1.0, 1.0], "RAS")


def is_voxel_sizes_valid(voxel_sizes):
    return is_reference_info_valid(np.eye(4), [1, 1, 1], voxel_sizes, "RAS")


def is_voxel_order_valid(voxel_order):
    return is_reference_info_valid(np.eye(4), [1, 1, 1], [1.0, 1.0, 1.0], voxel_order)


def test_reference_info_validity():
    assert_(not is_affine_valid(np.eye(3)), msg="3x3 affine is invalid")
    assert_(not is_affine_valid(np.zeros((4, 4))), msg="All zeroes affine is invalid")
    assert_(is_affine_valid(np.eye(4)), msg="Identity should be valid")

    assert_(not is_dimensions_valid([0, 0]), msg="Dimensions of the wrong length")
    assert_(not is_dimensions_valid([1, 1.0, 1]), msg="Dimensions cannot be float")
    assert_(not is_dimensions_valid([1, -1, 1]), msg="Dimensions cannot be negative")
    assert_(is_dimensions_valid([1, 1, 1]), msg="Dimensions of [1,1,1] should be valid")

    assert_(not is_voxel_sizes_valid([0, 0]), msg="Voxel sizes of the wrong length")
    assert_(not is_voxel_sizes_valid([1, -1, 1]), msg="Voxel sizes cannot be negative")
    assert_(
        is_voxel_sizes_valid([1.0, 1.0, 1.0]),
        msg="Voxel sizes of [1.0,1.0,1.0] should be valid",
    )

    assert_(not is_voxel_order_valid("RA"), msg="Voxel order of the wrong length")
    assert_(
        not is_voxel_order_valid(["RAS"]),
        msg="List of string is not a valid voxel order",
    )
    assert_(
        not is_voxel_order_valid(["R", "A", "Z"]),
        msg="Invalid value for voxel order (Z)",
    )
    assert_(not is_voxel_order_valid("RAZ"), msg="Invalid value for voxel order (Z)")
    assert_(is_voxel_order_valid("RAS"), msg="RAS should be a valid voxel order")
    assert_(
        is_voxel_order_valid(["R", "A", "S"]), msg="RAS should be a valid voxel order"
    )


def reference_info_zero_affine():
    header = create_nifti_header(np.zeros((4, 4)), [10, 10, 10], [1, 1, 1])
    try:
        get_reference_info(header)
        return True
    except ValueError:
        return False


def test_reference_trk_file_info_identical():
    tuple_1 = get_reference_info(FILEPATH_DIX["gs.trk"])
    tuple_2 = get_reference_info(FILEPATH_DIX["gs.nii"])
    affine_1, dimensions_1, voxel_sizes_1, voxel_order_1 = tuple_1
    affine_2, dimensions_2, voxel_sizes_2, voxel_order_2 = tuple_2

    assert_allclose(affine_1, affine_2)
    assert_array_equal(dimensions_1, dimensions_2)
    assert_allclose(voxel_sizes_1, voxel_sizes_2)
    assert voxel_order_1 == voxel_order_2


def test_reference_trx_file_info_identical():
    tuple_1 = get_reference_info(FILEPATH_DIX["gs.trx"])
    tuple_2 = get_reference_info(FILEPATH_DIX["gs.nii"])
    affine_1, dimensions_1, voxel_sizes_1, voxel_order_1 = tuple_1
    affine_2, dimensions_2, voxel_sizes_2, voxel_order_2 = tuple_2

    assert_allclose(affine_1, affine_2)
    assert_array_equal(dimensions_1, dimensions_2)
    assert_allclose(voxel_sizes_1, voxel_sizes_2)
    assert voxel_order_1 == voxel_order_2


def test_reference_obj_info_identical():
    sft = load_tractogram(FILEPATH_DIX["gs.trk"], "same")
    trx = tmm.load(FILEPATH_DIX["gs.trx"])
    img = nib.load(FILEPATH_DIX["gs.nii"])

    tuple_1 = get_reference_info(sft)
    tuple_2 = get_reference_info(trx)
    tuple_3 = get_reference_info(img)
    affine_1, dimensions_1, voxel_sizes_1, voxel_order_1 = tuple_1
    affine_2, dimensions_2, voxel_sizes_2, voxel_order_2 = tuple_2
    affine_3, dimensions_3, voxel_sizes_3, voxel_order_3 = tuple_3

    assert_allclose(affine_1, affine_2)
    assert_array_equal(dimensions_1, dimensions_2)
    assert_allclose(voxel_sizes_1, voxel_sizes_2)
    assert voxel_order_1 == voxel_order_2

    assert_allclose(affine_1, affine_3)
    assert_array_equal(dimensions_1, dimensions_3)
    assert_allclose(voxel_sizes_1, voxel_sizes_3)
    assert voxel_order_1 == voxel_order_3


def test_reference_header_info_identical():
    trk = nib.streamlines.load(FILEPATH_DIX["gs.trk"])
    trx = tmm.load(FILEPATH_DIX["gs.trx"])
    img = nib.load(FILEPATH_DIX["gs.nii"])

    tuple_1 = get_reference_info(trk.header)
    tuple_2 = get_reference_info(trx.header)
    tuple_3 = get_reference_info(img.header)
    affine_1, dimensions_1, voxel_sizes_1, voxel_order_1 = tuple_1
    affine_2, dimensions_2, voxel_sizes_2, voxel_order_2 = tuple_2
    affine_3, dimensions_3, voxel_sizes_3, voxel_order_3 = tuple_3

    assert_allclose(affine_1, affine_2)
    assert_array_equal(dimensions_1, dimensions_2)
    assert_allclose(voxel_sizes_1, voxel_sizes_2)
    assert voxel_order_1 == voxel_order_2

    assert_allclose(affine_1, affine_3)
    assert_array_equal(dimensions_1, dimensions_3)
    assert_allclose(voxel_sizes_1, voxel_sizes_3)
    assert voxel_order_1 == voxel_order_3


def test_all_zeros_affine():
    assert_(
        not reference_info_zero_affine(), msg="An all zeros affine should not be valid"
    )


@set_random_number_generator()
def test_read_img_arr_or_path(rng):
    data = rng.random((4, 4, 4, 3))
    aff = np.eye(4)
    aff[:3, :] = rng.standard_normal((3, 4))
    img = nib.Nifti1Image(data, aff)
    path = tempfile.NamedTemporaryFile().name + ".nii.gz"
    nib.save(img, path)
    for this in [data, img, path]:
        dd, aa = read_img_arr_or_path(this, affine=aff)
        assert np.allclose(dd, data)
        assert np.allclose(aa, aff)

    # Tests that if an array is provided, but no affine, an error is raised:
    with pytest.raises(ValueError):
        read_img_arr_or_path(data)

    # Tests that the affine is recovered correctly from path:
    dd, aa = read_img_arr_or_path(path)
    assert np.allclose(dd, data)
    assert np.allclose(aa, aff)



import numpy as np
import numpy.testing as npt

from dipy.reconst.utils import _mask_from_roi, _roi_in_volume, convert_tensors


def test_roi_in_volume():
    data_shape = (11, 11, 11, 64)
    roi_center = np.array([5, 5, 5])
    roi_radii = np.array([5, 5, 5])
    roi_radii_out = _roi_in_volume(data_shape, roi_center, roi_radii)
    npt.assert_array_equal(roi_radii_out, np.array([5, 5, 5]))

    roi_radii = np.array([6, 6, 6])
    roi_radii_out = _roi_in_volume(data_shape, roi_center, roi_radii)
    npt.assert_array_equal(roi_radii_out, np.array([5, 5, 5]))

    roi_center = np.array([4, 4, 4])
    roi_radii = np.array([5, 5, 5])
    roi_radii_out = _roi_in_volume(data_shape, roi_center, roi_radii)
    npt.assert_array_equal(roi_radii_out, np.array([4, 4, 4]))

    data_shape = (11, 11, 1, 64)
    roi_center = np.array([5, 5, 0])
    roi_radii = np.array([5, 5, 0])
    roi_radii_out = _roi_in_volume(data_shape, roi_center, roi_radii)
    npt.assert_array_equal(roi_radii_out, np.array([5, 5, 0]))

    roi_center = np.array([2, 5, 0])
    roi_radii = np.array([5, 10, 2])
    roi_radii_out = _roi_in_volume(data_shape, roi_center, roi_radii)
    npt.assert_array_equal(roi_radii_out, np.array([2, 5, 0]))


def test_mask_from_roi():
    data_shape = (5, 5, 5)
    roi_center = (2, 2, 2)
    roi_radii = (2, 2, 2)
    mask_gt = np.ones(data_shape)
    roi_mask = _mask_from_roi(data_shape, roi_center, roi_radii)
    npt.assert_array_equal(roi_mask, mask_gt)

    roi_radii = (1, 2, 2)
    mask_gt = np.zeros(data_shape)
    mask_gt[1:4, 0:5, 0:5] = 1
    roi_mask = _mask_from_roi(data_shape, roi_center, roi_radii)
    npt.assert_array_equal(roi_mask, mask_gt)

    roi_radii = (0, 2, 2)
    mask_gt = np.zeros(data_shape)
    mask_gt[2, 0:5, 0:5] = 1
    roi_mask = _mask_from_roi(data_shape, roi_center, roi_radii)
    npt.assert_array_equal(roi_mask, mask_gt)


def test_convert_tensor():
    # Test case 1: Convert from 'dipy' to 'mrtrix'
    tensor = np.array([[[[1, 2, 3, 4, 5, 6]]]])
    converted_tensor = convert_tensors(tensor, "dipy", "mrtrix")
    expected_tensor = np.array([[[[1, 3, 6, 2, 4, 5]]]])
    npt.assert_array_equal(converted_tensor, expected_tensor)

    # Test case 2: Convert from 'mrtrix' to 'ants'
    tensor = np.array([[[[1, 2, 3, 4, 5, 6]]]])
    converted_tensor = convert_tensors(tensor, "mrtrix", "ants")
    expected_tensor = np.array([[[[[1, 4, 2, 5, 6, 3]]]]])
    npt.assert_array_equal(converted_tensor, expected_tensor)

    # Test case 3: Convert from 'ants' to 'fsl'
    tensor = np.array([[[[1, 2, 3, 4, 5, 6]]]])
    converted_tensor = convert_tensors(tensor, "ants", "fsl")
    expected_tensor = np.array([[[[1, 2, 4, 3, 5, 6]]]])
    npt.assert_array_equal(converted_tensor, expected_tensor)

    # Test case 4: Convert from 'fsl' to 'dipy'
    tensor = np.array([[[[1, 2, 3, 4, 5, 6]]]])
    converted_tensor = convert_tensors(tensor, "fsl", "dipy")
    expected_tensor = np.array([[[[1, 2, 4, 3, 5, 6]]]])
    npt.assert_array_equal(converted_tensor, expected_tensor)

    # Test case 5: Convert from 'dipy' to 'ants'
    tensor = np.array([[[[1, 2, 3, 4, 5, 6]]]])
    converted_tensor = convert_tensors(tensor, "dipy", "ants")
    expected_tensor = np.array([[[[[1, 2, 3, 4, 5, 6]]]]])
    npt.assert_array_equal(converted_tensor, expected_tensor)

    # Test case 6: Convert from 'ants' to 'dipy'
    tensor = np.array([[[[[1, 2, 3, 4, 5, 6]]]]])
    converted_tensor = convert_tensors(tensor, "ants", "dipy")
    expected_tensor = np.array([1, 2, 3, 4, 5, 6])
    npt.assert_array_equal(converted_tensor, expected_tensor)

    # Test case 7: Convert from 'dipy' to 'dipy'
    tensor = np.array([[[[[1, 2, 3, 4, 5, 6]]]]])
    converted_tensor = convert_tensors(tensor, "dipy", "dipy")
    npt.assert_array_equal(converted_tensor, tensor)

    npt.assert_raises(ValueError, convert_tensors, tensor, "amico", "dipy")
    npt.assert_raises(ValueError, convert_tensors, tensor, "dipy", "amico")



"""
==================================
BUAN Bundle Shape Similarity Score
==================================

This example explains how we can use BUAN :footcite:p:`Chandio2020a` to calculate
shape similarity between two given bundles. Where, shape similarity score of 1
means two bundles are extremely close in shape and 0 implies no shape similarity
whatsoever.

Shape similarity score can be used to compare populations or individuals.
It can also serve as a quality assurance metric, to validate streamline
registration quality, bundle extraction quality by calculating output with a
reference bundle or other issues with pre-processing by calculating shape
dissimilarity with a reference bundle.

First import the necessary modules.
"""

import numpy as np

from dipy.data import two_cingulum_bundles
from dipy.segment.bundles import (
    bundle_shape_similarity,
    select_random_set_of_streamlines,
)
from dipy.viz import actor, window

###############################################################################
# To show the concept we will use two pre-saved cingulum bundle.
# Let's start by fetching the data.

cb_subj1, _ = two_cingulum_bundles()

###############################################################################
# Let's create two streamline sets (bundles) from same bundle cb_subj1 by
# randomly selecting 60 streamlines two times.

rng = np.random.default_rng()
bundle1 = select_random_set_of_streamlines(cb_subj1, 60, rng=None)
bundle2 = select_random_set_of_streamlines(cb_subj1, 60, rng=None)

###############################################################################
# Now, let's visualize two bundles.


def show_both_bundles(bundles, colors=None, show=True, fname=None):
    scene = window.Scene()
    scene.SetBackground(1.0, 1, 1)
    for i, bundle in enumerate(bundles):
        color = colors[i]
        streamtube_actor = actor.streamtube(bundle, colors=color, linewidth=0.3)
        streamtube_actor.RotateX(-90)
        streamtube_actor.RotateZ(90)
        scene.add(streamtube_actor)
    if show:
        window.show(scene)
    if fname is not None:
        window.record(scene=scene, n_frames=1, out_path=fname, size=(900, 900))


show_both_bundles(
    [bundle1, bundle2],
    colors=[(1, 0, 0), (0, 1, 0)],
    show=False,
    fname="two_bundles.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Two Cingulum Bundles.
#
#
#
# Calculate shape similarity score between two bundles.
# 0 cluster_thr because we want to use all streamlines and not the centroids of
# clusters.

clust_thr = [0]

###############################################################################
# Threshold indicates how strictly we want two bundles to be similar in shape.

threshold = 5

ba_score = bundle_shape_similarity(
    bundle1, bundle2, rng, clust_thr=clust_thr, threshold=threshold
)
print("Shape similarity score = ", ba_score)

###############################################################################
# Let's change the value of threshold to 10.

threshold = 10

ba_score = bundle_shape_similarity(
    bundle1, bundle2, rng, clust_thr=clust_thr, threshold=threshold
)
print("Shape similarity score = ", ba_score)

###############################################################################
# Higher value of threshold gives us higher shape similarity score as it is
# more lenient.
#
#
#
# References
# ----------
#
# .. footbibliography::
#



import warnings

import numpy as np
import numpy.testing as npt

from dipy.core.geometry import cart2sphere
from dipy.core.gradients import gradient_table
from dipy.core.sphere import HemiSphere, unit_icosahedron
from dipy.data import get_fnames, get_sphere
from dipy.direction.bootstrap_direction_getter import BootDirectionGetter
from dipy.io.gradients import read_bvals_bvecs
from dipy.reconst import dti, shm
from dipy.reconst.csdeconv import ConstrainedSphericalDeconvModel, TensorModel
from dipy.sims.voxel import multi_tensor, single_tensor
from dipy.testing.decorators import set_random_number_generator


def test_bdg_initial_direction():
    """This tests the number of initial directions." """

    hsph_updated = HemiSphere.from_sphere(unit_icosahedron).subdivide(n=2)
    vertices = hsph_updated.vertices
    bvecs = vertices
    bvals = np.ones(len(vertices)) * 1000
    bvecs = np.insert(bvecs, 0, np.array([0, 0, 0]), axis=0)
    bvals = np.insert(bvals, 0, 0)
    gtab = gradient_table(bvals, bvecs=bvecs)

    # test that we get one direction when we have a single tensor
    sphere = HemiSphere.from_sphere(get_sphere(name="symmetric724"))
    voxel = single_tensor(gtab).reshape([1, 1, 1, -1])
    dti_model = dti.TensorModel(gtab)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        boot_dg = BootDirectionGetter.from_data(
            voxel, dti_model, 30, sphere=sphere, sh_order=6
        )
    initial_direction = boot_dg.initial_direction(np.zeros(3))
    npt.assert_equal(len(initial_direction), 1)
    npt.assert_allclose(initial_direction[0], [1, 0, 0], atol=0.1)

    # test that we get multiple directions when we have a multi-tensor
    mevals = np.array([[1.5, 0.4, 0.4], [1.5, 0.4, 0.4]]) * 1e-3
    fracs = [60, 40]
    voxel, primary_evecs = multi_tensor(gtab, mevals, fractions=fracs, snr=None)
    voxel = voxel.reshape([1, 1, 1, -1])
    response = (np.array([0.0015, 0.0004, 0.0004]), 1)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        csd_model = ConstrainedSphericalDeconvModel(
            gtab, response=response, sh_order_max=4
        )
        boot_dg = BootDirectionGetter.from_data(
            voxel,
            csd_model,
            30,
            sphere=sphere,
        )
        initial_direction = boot_dg.initial_direction(np.zeros(3))

    npt.assert_equal(len(initial_direction), 2)
    npt.assert_allclose(initial_direction, primary_evecs, atol=0.1)


def test_bdg_get_direction():
    """This tests the direction found by the bootstrap direction getter."""

    _, fbvals, fbvecs = get_fnames(name="small_64D")

    bvals, bvecs = read_bvals_bvecs(fbvals, fbvecs)
    gtab = gradient_table(bvals, bvecs=bvecs, b0_threshold=0)
    mevals = np.array(([0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]))

    angles = [(0, 0)]

    voxel, _ = multi_tensor(gtab, mevals, S0=1, angles=angles, fractions=[100], snr=100)
    data = np.tile(voxel, (3, 3, 3, 1))
    sphere = get_sphere(name="symmetric362")
    response = (np.array([0.0015, 0.0003, 0.0003]), 1)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        csd_model = ConstrainedSphericalDeconvModel(gtab, response, sh_order_max=6)

    point = np.array([0.0, 0.0, 0.0])
    prev_direction = sphere.vertices[5]

    # test case in which no valid direction is found with default max attempts
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        boot_dg = BootDirectionGetter(
            data, model=csd_model, max_angle=10.0, sphere=sphere
        )
        npt.assert_equal(boot_dg.get_direction(point, prev_direction), 1)

    # test case in which no valid direction is found with new max attempts
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        boot_dg = BootDirectionGetter(
            data, model=csd_model, max_angle=10, sphere=sphere, max_attempts=3
        )
        npt.assert_equal(boot_dg.get_direction(point, prev_direction), 1)

    # test case in which a valid direction is found
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        boot_dg = BootDirectionGetter(
            data, model=csd_model, max_angle=60.0, sphere=sphere, max_attempts=5
        )
        npt.assert_equal(boot_dg.get_direction(point, prev_direction), 0)

    # test invalid max_attempts parameters
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        npt.assert_raises(
            ValueError,
            lambda: BootDirectionGetter(
                data, csd_model, 60, sphere=sphere, max_attempts=0
            ),
        )


@set_random_number_generator()
def test_bdg_residual(rng):
    """This tests the bootstrapping residual."""

    hsph_updated = HemiSphere.from_sphere(unit_icosahedron).subdivide(n=2)
    vertices = hsph_updated.vertices
    bvecs = vertices
    bvals = np.ones(len(vertices)) * 1000
    bvecs = np.insert(bvecs, 0, np.array([0, 0, 0]), axis=0)
    bvals = np.insert(bvals, 0, 0)
    gtab = gradient_table(bvals, bvecs=bvecs)
    r, theta, phi = cart2sphere(*vertices.T)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        B, m, n = shm.real_sh_descoteaux(6, theta, phi)
    shm_coeff = rng.random(B.shape[1])

    # sphere_func is sampled of the spherical function for each point of
    # the sphere
    sphere_func = np.dot(shm_coeff, B.T)

    response = (np.array([1.5e3, 0.3e3, 0.3e3]), 1)
    voxel = np.concatenate((np.zeros(1), sphere_func))
    data = np.tile(voxel, (3, 3, 3, 1))

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        csd_model = ConstrainedSphericalDeconvModel(gtab, response, sh_order_max=6)
        boot_dg = BootDirectionGetter.from_data(
            data, model=csd_model, max_angle=60, sphere=hsph_updated, sh_order=6
        )

        # Two boot samples should be the same
        odf1 = boot_dg.get_pmf(np.array([1.5, 1.5, 1.5]))
        odf2 = boot_dg.get_pmf(np.array([1.5, 1.5, 1.5]))
        npt.assert_array_almost_equal(odf1, odf2)

    # A boot sample with less sh coeffs should have residuals, thus the two
    # should be different
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        boot_dg2 = BootDirectionGetter.from_data(
            data, model=csd_model, max_angle=60, sphere=hsph_updated, sh_order=4
        )
    odf1 = boot_dg2.get_pmf(np.array([1.5, 1.5, 1.5]))
    odf2 = boot_dg2.get_pmf(np.array([1.5, 1.5, 1.5]))
    npt.assert_(np.any(odf1 != odf2))

    # test with a gtab with two shells and assert you get an error
    bvals[-1] = 2000
    gtab = gradient_table(bvals, bvecs=bvecs)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        csd_model = ConstrainedSphericalDeconvModel(gtab, response, sh_order_max=6)
    npt.assert_raises(
        ValueError,
        BootDirectionGetter,
        data,
        csd_model,
        60,
        sphere=hsph_updated,
        max_attempts=6,
    )


def test_boot_pmf():
    # This tests the local model used for the bootstrapping.
    hsph_updated = HemiSphere.from_sphere(unit_icosahedron)
    vertices = hsph_updated.vertices
    bvecs = vertices
    bvals = np.ones(len(vertices)) * 1000
    bvecs = np.insert(bvecs, 0, np.array([0, 0, 0]), axis=0)
    bvals = np.insert(bvals, 0, 0)
    gtab = gradient_table(bvals, bvecs=bvecs)
    voxel = single_tensor(gtab)
    data = np.tile(voxel, (3, 3, 3, 1))
    point = np.array([1.0, 1.0, 1.0])
    tensor_model = TensorModel(gtab)
    response = (np.array([1.5e3, 0.3e3, 0.3e3]), 1)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        boot_dg = BootDirectionGetter(
            data, model=tensor_model, max_angle=60, sphere=hsph_updated
        )
    no_boot_pmf = boot_dg.get_pmf_no_boot(point)
    model_pmf = tensor_model.fit(voxel).odf(hsph_updated)

    npt.assert_equal(len(hsph_updated.vertices), no_boot_pmf.shape[0])
    npt.assert_array_almost_equal(no_boot_pmf, model_pmf)

    # test model spherical harmonic order different than bootstrap order
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always", category=UserWarning)
        warnings.simplefilter("always", category=PendingDeprecationWarning)
        csd_model = ConstrainedSphericalDeconvModel(gtab, response, sh_order_max=6)
        # Tests that the first caught warning comes from the CSD model
        # constructor
        npt.assert_(issubclass(w[0].category, UserWarning))
        npt.assert_("Number of parameters required " in str(w[0].message))

        # Tests that additional warnings are raised for outdated SH basis
        npt.assert_(len(w) > 1)

        boot_dg_sh4 = BootDirectionGetter(
            data, model=csd_model, max_angle=60, sphere=hsph_updated, sh_order=4
        )
        pmf_sh4 = boot_dg_sh4.get_pmf(point)
        npt.assert_equal(len(hsph_updated.vertices), pmf_sh4.shape[0])
        npt.assert_(np.sum(pmf_sh4.shape) > 0)

        boot_dg_sh8 = BootDirectionGetter(
            data, model=csd_model, max_angle=60, sphere=hsph_updated, sh_order=8
        )
        pmf_sh8 = boot_dg_sh8.get_pmf(point)
        npt.assert_equal(len(hsph_updated.vertices), pmf_sh8.shape[0])
        npt.assert_(np.sum(pmf_sh8.shape) > 0)

    # test b_tol parameter
    bvals[-2] = 1100
    gtab = gradient_table(bvals, bvecs=bvecs)
    tensor_model = TensorModel(gtab)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        npt.assert_raises(
            ValueError,
            BootDirectionGetter,
            data,
            tensor_model,
            60,
            sphere=hsph_updated,
            max_attempts=6,
            b_tol=20,
        )
        npt.assert_raises(
            ValueError,
            BootDirectionGetter,
            data,
            tensor_model,
            60,
            sphere=hsph_updated,
            max_attempts=6,
            b_tol=-1,
        )



"""
============================================================================
Tracking with Robust Unbiased Model-BAsed Spherical Deconvolution (RUMBA-SD)
============================================================================

Here, we demonstrate fiber tracking using a probabilistic direction getter
and RUMBA-SD, a model introduced in :footcite:p:`CanalesRodriguez2015`. This
model adapts Richardson-Lucy deconvolution by assuming Rician or Noncentral Chi
noise instead of Gaussian, which more accurately reflects the noise from MRI
scanners (see also
:ref:`sphx_glr_examples_built_reconstruction_reconst_rumba.py`). This tracking
tutorial is an extension on
:ref:`sphx_glr_examples_built_fiber_tracking_tracking_probabilistic.py`.

We start by loading sample data and identifying a fiber response function.
"""

import matplotlib.pyplot as plt
from numpy.linalg import inv

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames, small_sphere
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_trk
from dipy.reconst.csdeconv import auto_response_ssst
from dipy.reconst.rumba import RumbaSDModel
from dipy.tracking import utils
from dipy.tracking.stopping_criterion import ThresholdStoppingCriterion
from dipy.tracking.streamline import Streamlines, transform_streamlines
from dipy.tracking.tracker import probabilistic_tracking
from dipy.viz import actor, colormap, window

# Enables/disables interactive visualization
interactive = False

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")
t1_fname = get_fnames(name="stanford_t1")

data, affine, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)
t1_data, t1_aff = load_nifti(t1_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

seed_mask = labels == 2
white_matter = (labels == 1) | (labels == 2)
seeds = utils.seeds_from_mask(seed_mask, affine, density=2)

response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)

sphere = small_sphere

###############################################################################
# We can now initialize a `RumbaSdModel` model and fit it globally by setting
# `voxelwise` to `False`. For this example, TV regularization (`use_tv`) will
# be turned off for efficiency, although its usage can provide more coherent
# results in fiber tracking. The fit will take about 5 minutes to complete.

rumba = RumbaSDModel(
    gtab,
    wm_response=response[0],
    n_iter=200,
    voxelwise=False,
    use_tv=False,
    sphere=sphere,
)
rumba_fit = rumba.fit(data, mask=white_matter)
odf = rumba_fit.odf()  # fODF
f_wm = rumba_fit.f_wm  # white matter volume fractions

###############################################################################
# To establish stopping criterion, a common technique is to use the Generalized
# Fractional Anisotropy (GFA). One point of caution is that RUMBA-SD by default
# separates the fODF from an isotropic compartment. This can bias GFA results
# computed on the fODF, although it will still generally be an effective
# criterion.
#
# However, an alternative stopping criterion that takes advantage of this
# feature is to use RUMBA-SD's white matter volume fraction map.

stopping_criterion = ThresholdStoppingCriterion(f_wm, 0.25)

###############################################################################
# We can visualize a slice of this mask.

sli = f_wm.shape[2] // 2
plt.figure()

plt.subplot(1, 2, 1).set_axis_off()
plt.imshow(f_wm[:, :, sli].T, cmap="gray", origin="lower")

plt.subplot(1, 2, 2).set_axis_off()
plt.imshow((f_wm[:, :, sli] > 0.25).T, cmap="gray", origin="lower")

plt.savefig("f_wm_tracking_mask.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# White matter volume fraction slice
#
#
#
# These discrete fODFs can be used as a PMF in the
# `ProbabilisticDirectionGetter` for sampling tracking directions. The PMF
# must be strictly non-negative; RUMBA-SD already adheres to this constraint
# so no further manipulation of the fODFs is necessary.

streamline_generator = probabilistic_tracking(
    seeds,
    stopping_criterion,
    affine,
    step_size=0.5,
    max_angle=30.0,
    sphere=sphere,
    sf=odf,
)
streamlines = Streamlines(streamline_generator)

color = colormap.line_colors(streamlines)
streamlines_actor = actor.streamtube(
    list(transform_streamlines(streamlines, inv(t1_aff))), colors=color, linewidth=0.1
)

vol_actor = actor.slicer(t1_data)
vol_actor.display(x=40)
vol_actor2 = vol_actor.copy()
vol_actor2.display(z=35)

scene = window.Scene()
scene.add(vol_actor)
scene.add(vol_actor2)
scene.add(streamlines_actor)
if interactive:
    window.show(scene)

window.record(
    scene=scene, out_path="tractogram_probabilistic_rumba.png", size=(800, 800)
)

sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_probabilistic_rumba.trk")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# RUMBA-SD tractogram
#
#
#
# References
# ----------
#
# .. footbibliography::
#



"""
==============================================
Denoise images using Non-Local Means (NLMEANS)
==============================================

Using the non-local means filter :footcite:p:`Coupe2008` and
:footcite:p:`Coupe2012` and you can denoise 3D or 4D images and boost the SNR of
your datasets. You can also decide between modeling the noise as Gaussian or
Rician (default).

We start by loading the necessary modules
"""

from time import time

import matplotlib.pyplot as plt
import numpy as np

from dipy.data import get_fnames
from dipy.denoise.nlmeans import nlmeans
from dipy.denoise.noise_estimate import estimate_sigma
from dipy.io.image import load_nifti, save_nifti

###############################################################################
# Then, let's fetch and load a T1 data from Stanford University

t1_fname = get_fnames(name="stanford_t1")
data, affine = load_nifti(t1_fname)

mask = data > 1500

print("vol size", data.shape)

###############################################################################
# In order to call ``non_local_means`` first you need to estimate the standard
# deviation of the noise. We have used N=32 since the Stanford dataset was
# acquired on a 3T GE scanner with a 32 array head coil.

sigma = estimate_sigma(data, N=32)

###############################################################################
# Calling the main function ``non_local_means``

t = time()

den = nlmeans(data, sigma=sigma, mask=mask, patch_radius=1, block_radius=2, rician=True)

print("total time", time() - t)

###############################################################################
# Let us plot the axial slice of the denoised output

axial_middle = data.shape[2] // 2

before = data[:, :, axial_middle].T
after = den[:, :, axial_middle].T

difference = np.abs(after.astype(np.float64) - before.astype(np.float64))

difference[~mask[:, :, axial_middle].T] = 0


fig, ax = plt.subplots(1, 3)
ax[0].imshow(before, cmap="gray", origin="lower")
ax[0].set_title("before")
ax[1].imshow(after, cmap="gray", origin="lower")
ax[1].set_title("after")
ax[2].imshow(difference, cmap="gray", origin="lower")
ax[2].set_title("difference")

plt.savefig("denoised.png", bbox_inches="tight")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Showing axial slice before (left) and after (right) NLMEANS denoising

save_nifti("denoised.nii.gz", den, affine)

###############################################################################
# An improved version of non-local means denoising is adaptive soft coefficient
# matching, please refer to
# :ref:`sphx_glr_examples_built_preprocessing_denoise_ascm.py` for more
# details.
#
# References
# ----------
#
# .. footbibliography::
#



# cython: wraparound=False, cdivision=True, boundscheck=False

cimport numpy as cnp
import numpy as np

from dipy.core.interpolation cimport trilinear_interpolate4d_c
from dipy.data import default_sphere
from dipy.direction.closest_peak_direction_getter cimport closest_peak
from dipy.direction.peaks import peak_directions
from dipy.reconst import shm
from dipy.tracking.direction_getter cimport DirectionGetter


cdef class BootDirectionGetter(DirectionGetter):

    cdef:
        cnp.ndarray dwi_mask
        double[:] vox_data
        dict _pf_kwargs
        double cos_similarity
        double min_separation_angle
        double relative_peak_threshold
        double[:] pmf
        double[:, :] R
        double[:, :, :, :] data
        int max_attempts
        int sh_order
        object H
        object model
        object sphere


    def __init__(self, data, model, max_angle, sphere=default_sphere,
                 max_attempts=5, sh_order=0, b_tol=20, **kwargs):
        cdef:
            cnp.ndarray x, y, z, r
            double[:] theta, phi
            double[:, :] B

        if max_attempts < 1:
             raise ValueError("max_attempts must be greater than 0.")

        if b_tol <= 0:
            raise ValueError("b_tol must be greater than 0.")

        self._pf_kwargs = kwargs
        self.data = np.asarray(data, dtype=float)
        self.model = model
        self.cos_similarity = np.cos(np.deg2rad(max_angle))
        self.sphere = sphere
        self.sh_order = sh_order
        self.max_attempts = max_attempts

        if self.sh_order == 0:
            if hasattr(model, "sh_order"):
                self.sh_order = model.sh_order
            else:
                self.sh_order = 4 #  DEFAULT Value

        self.dwi_mask = model.gtab.b0s_mask == 0
        x, y, z = model.gtab.gradients[self.dwi_mask].T
        r, theta, phi = shm.cart2sphere(x, y, z)
        if r.max() - r.min() >= b_tol:
            raise ValueError("BootDirectionGetter only supports single shell \
                              data.")
        B, _, _ = shm.real_sh_descoteaux(self.sh_order, theta, phi)
        self.H = shm.hat(B)
        self.R = shm.lcr_matrix(self.H)

        self.vox_data = np.empty(self.data.shape[3])
        self.pmf = np.empty(sphere.vertices.shape[0])


    @classmethod
    def from_data(cls, data, model, max_angle, sphere=default_sphere,
                  sh_order=0, max_attempts=5, b_tol=20, **kwargs):
        """Create a BootDirectionGetter using HARDI data and an ODF type model

        Parameters
        ----------
        data : ndarray, float, (..., N)
            Diffusion MRI data with N volumes.
        model : dipy diffusion model
            Must provide fit with odf method.
        max_angle : float (0, 90)
            Maximum angle between tract segments. This angle can be more
            generous (larger) than values typically used with probabilistic
            direction getters.
        sphere : Sphere
            The sphere used to sample the diffusion ODF.
        sh_order : even int
            The order of the SH "model" used to estimate bootstrap residuals.
        max_attempts : int
            Max number of bootstrap samples used to find tracking direction
            before giving up.
        b_tol : float
            Maximum difference between b-values to be considered single shell.
        relative_peak_threshold : float in [0., 1.]
            Relative threshold for excluding ODF peaks.
        min_separation_angle : float in [0, 90]
            Angular threshold for excluding ODF peaks.

        """
        return cls(data, model, max_angle, sphere=sphere, max_attempts=max_attempts, sh_order=sh_order,
                   b_tol=b_tol, **kwargs)


    cpdef cnp.ndarray[cnp.float_t, ndim=2] initial_direction(self,
                                                             double[::1] point):
        """Returns best directions at seed location to start tracking.

        Parameters
        ----------
        point : ndarray, shape (3,)
            The point in an image at which to lookup tracking directions.

        Returns
        -------
        directions : ndarray, shape (N, 3)
            Possible tracking directions from point. ``N`` may be 0, all
            directions should be unique.

        """
        cdef:
            double[:] pmf = self.get_pmf_no_boot(point)

        return peak_directions(pmf, self.sphere, **self._pf_kwargs)[0]


    cpdef double[:] get_pmf(self, double[::1] point):
        """Produces an ODF from a SH bootstrap sample"""
        if trilinear_interpolate4d_c(self.data,
                                     &point[0],
                                     &self.vox_data[0]) != 0:
            self.__clear_pmf()
        else:
            np.asarray(self.vox_data)[self.dwi_mask] = shm.bootstrap_data_voxel(
                np.asarray(self.vox_data)[self.dwi_mask], self.H, self.R)
            self.pmf = self.model.fit(np.asarray(self.vox_data)).odf(self.sphere)
        return self.pmf


    cpdef double[:] get_pmf_no_boot(self, double[::1] point):

        if trilinear_interpolate4d_c(self.data,
                                     &point[0],
                                     &self.vox_data[0]) != 0:
            self.__clear_pmf()
        else:
            self.pmf = self.model.fit(np.asarray(self.vox_data)).odf(self.sphere)
        return self.pmf

    cdef void __clear_pmf(self) nogil:
        cdef:
            cnp.npy_intp len_pmf = self.pmf.shape[0]
            cnp.npy_intp i

        for i in range(len_pmf):
            self.pmf[i] = 0.0


    cdef int get_direction_c(self, double[::1] point, double[::1] direction):
        """Attempt direction getting on a few bootstrap samples.

        Returns
        -------
        status : int
            Returns 0 `direction` was updated with a new tracking direction, or
            1 otherwise.
        """
        cdef:
            double[:] pmf
            cnp.ndarray[cnp.float_t, ndim=2] peaks

        for _ in range(self.max_attempts):
            pmf = self.get_pmf(point)
            peaks = peak_directions(pmf, self.sphere, **self._pf_kwargs)[0]
            if len(peaks) > 0:
                return closest_peak(peaks, direction, self.cos_similarity)
        return 1



from .bootstrap_direction_getter import BootDirectionGetter
from .closest_peak_direction_getter import ClosestPeakDirectionGetter
from .peaks import (
    PeaksAndMetrics,
    peak_directions,
    peak_directions_nl,
    peaks_from_model,
    peaks_from_positions,
    reshape_peaks_for_visualization,
)
from .probabilistic_direction_getter import (
    DeterministicMaximumDirectionGetter,
    ProbabilisticDirectionGetter,
)
from .ptt_direction_getter import PTTDirectionGetter

__all__ = [
    "BootDirectionGetter",
    "ClosestPeakDirectionGetter",
    "DeterministicMaximumDirectionGetter",
    "ProbabilisticDirectionGetter",
    "PTTDirectionGetter",
    "PeaksAndMetrics",
    "peak_directions",
    "peak_directions_nl",
    "peaks_from_model",
    "peaks_from_positions",
    "reshape_peaks_for_visualization",
]



from collections.abc import Iterable
import random
from warnings import warn

import numpy as np

from dipy.testing.decorators import warning_for_keywords
from dipy.tracking import utils
from dipy.tracking.localtrack import local_tracker, pft_tracker
from dipy.tracking.stopping_criterion import (
    AnatomicalStoppingCriterion,
    StreamlineStatus,
)
from dipy.utils import fast_numpy


class LocalTracking:
    @staticmethod
    def _get_voxel_size(affine):
        """Computes the voxel sizes of an image from the affine.

        Checks that the affine does not have any shear because local_tracker
        assumes that the data is sampled on a regular grid.

        """
        lin = affine[:3, :3]
        dotlin = np.dot(lin.T, lin)
        # Check that the affine is well behaved
        if not np.allclose(np.triu(dotlin, 1), 0.0, atol=1e-5):
            msg = (
                "The affine provided seems to contain shearing, data must "
                "be acquired or interpolated on a regular grid to be used "
                "with `LocalTracking`."
            )
            raise ValueError(msg)
        return np.sqrt(dotlin.diagonal())

    @warning_for_keywords()
    def __init__(
        self,
        direction_getter,
        stopping_criterion,
        seeds,
        affine,
        step_size,
        *,
        max_cross=None,
        maxlen=500,
        minlen=2,
        fixedstep=True,
        return_all=True,
        random_seed=None,
        save_seeds=False,
        unidirectional=False,
        randomize_forward_direction=False,
        initial_directions=None,
    ):
        """Creates streamlines by using local fiber-tracking.

        Parameters
        ----------
        direction_getter : instance of DirectionGetter
            Used to get directions for fiber tracking.
        stopping_criterion : instance of StoppingCriterion
            Identifies endpoints and invalid points to inform tracking.
        seeds : array (N, 3), optional
            Points to seed the tracking. Seed points should be given in point
            space of the track (see ``affine``).
        affine : array (4, 4), optional
            Coordinate space for the streamline point with respect to voxel
            indices of input data. This affine can contain scaling, rotational,
            and translational components but should not contain any shearing.
            An identity matrix can be used to generate streamlines in "voxel
            coordinates" as long as isotropic voxels were used to acquire the
            data.
        step_size : float, optional
            Step size used for tracking.
        max_cross : int or None, optional
            The maximum number of direction to track from each seed in crossing
            voxels. By default all initial directions are tracked.
        maxlen : int, optional
            Maximum length of generated streamlines. Longer streamlines will be
            discarted if `return_all=False`.
        minlen : int, optional
            Minimum length of generated streamlines. Shorter streamlines will
            be discarted if `return_all=False`.
        fixedstep : bool, optional
            If true, a fixed stepsize is used, otherwise a variable step size
            is used.
        return_all : bool, optional
            If true, return all generated streamlines, otherwise only
            streamlines reaching end points or exiting the image.
        random_seed : int, optional
            The seed for the random seed generator (numpy.random.seed and
            random.seed).
        save_seeds : bool, optional
            If True, return seeds alongside streamlines
        unidirectional : bool, optional
            If true, the tracking is performed only in the forward direction.
            The seed position will be the first point of all streamlines.
        randomize_forward_direction : bool, optional
            If true, the forward direction is randomized (multiplied by 1
            or -1). Otherwise, the provided forward direction is used.
        initial_directions: array (N, npeaks, 3), optional
            Initial direction to follow from the ``seed`` position. If
            ``max_cross`` is None, one streamline will be generated per peak
            per voxel. If None, `direction_getter.initial_direction` is used.
        """

        self.direction_getter = direction_getter
        self.stopping_criterion = stopping_criterion
        self.seeds = seeds
        self.unidirectional = unidirectional
        self.randomize_forward_direction = randomize_forward_direction
        self.initial_directions = initial_directions

        if affine.shape != (4, 4):
            raise ValueError("affine should be a (4, 4) array.")
        if step_size <= 0:
            raise ValueError("step_size must be greater than 0.")
        if maxlen < 1:
            raise ValueError("maxlen must be greater than 0.")
        if minlen > maxlen:
            raise ValueError("maxlen must be greater than or equal to minlen")
        if not isinstance(seeds, Iterable):
            raise ValueError("seeds should be (N,3) array.")
        if (
            initial_directions is not None
            and seeds.shape[0] != initial_directions.shape[0]
        ):
            raise ValueError(
                "initial_directions and seeds must have the same shape[0]."
            )
        if (
            initial_directions is None
            and unidirectional
            and not self.randomize_forward_direction
        ):
            warn(
                "Unidirectional tractography will be performed "
                + "without providing initial directions nor "
                + "randomizing extracted initial forward "
                + "directions. This may introduce directional "
                + "biases in the reconstructed streamlines. "
                + "See ``initial_directions`` and "
                + "``randomize_forward_direction`` parameters.",
                stacklevel=2,
            )

        self.affine = affine
        self._voxel_size = np.ascontiguousarray(
            self._get_voxel_size(affine), dtype=float
        )
        self.step_size = step_size
        self.fixed_stepsize = fixedstep
        self.max_cross = max_cross
        self.max_length = maxlen
        self.min_length = minlen
        self.return_all = return_all
        self.random_seed = random_seed
        self.save_seeds = save_seeds

    def _tracker(self, seed, first_step, streamline):
        return local_tracker(
            self.direction_getter,
            self.stopping_criterion,
            seed,
            first_step,
            self._voxel_size,
            streamline,
            self.step_size,
            self.fixed_stepsize,
        )

    def __iter__(self):
        # Make tracks, move them to point space and return
        track = self._generate_tractogram()

        return utils.transform_tracking_output(
            track, self.affine, save_seeds=self.save_seeds
        )

    def _generate_tractogram(self):
        """A streamline generator"""

        # Get inverse transform (lin/offset) for seeds
        inv_A = np.linalg.inv(self.affine)
        lin = inv_A[:3, :3]
        offset = inv_A[:3, 3]

        F = np.empty((self.max_length + 1, 3), dtype=float)
        B = F.copy()
        for i, s in enumerate(self.seeds):
            s = np.dot(lin, s) + offset
            # Set the random seed in numpy, random and fast_numpy (lic.stdlib)
            if self.random_seed is not None:
                s_random_seed = hash(np.abs((np.sum(s)) + self.random_seed)) % (
                    np.iinfo(np.uint32).max - 1
                )
                random.seed(s_random_seed)
                np.random.seed(s_random_seed)
                fast_numpy.seed(s_random_seed)

            if self.initial_directions is None:
                directions = self.direction_getter.initial_direction(s)
            else:
                # normalize the initial directions.
                # initial directions with norm 0 are removed.
                d_ns = np.linalg.norm(self.initial_directions[i, :, :], axis=1)
                directions = (
                    self.initial_directions[i, d_ns > 0, :] / d_ns[d_ns > 0, np.newaxis]
                )

            if len(directions) == 0 and self.return_all:
                # only the seed position
                if self.save_seeds:
                    yield [s], s
                else:
                    yield [s]

            if self.randomize_forward_direction:
                directions = [d * random.choice([1, -1]) for d in directions]

            directions = directions[: self.max_cross]

            for first_step in directions:
                stepsF = stepsB = 1
                stepsF, stream_status = self._tracker(s, first_step, F)
                if not (
                    self.return_all
                    or stream_status
                    in (StreamlineStatus.ENDPOINT, StreamlineStatus.OUTSIDEIMAGE)
                ):
                    continue

                if not self.unidirectional:
                    first_step = -first_step
                    if stepsF > 1:
                        # Use the opposite of the first selected orientation for
                        # the backward tracking segment
                        opposite_step = F[0] - F[1]
                        opposite_step_norm = np.linalg.norm(opposite_step)
                        if opposite_step_norm > 0:
                            first_step = opposite_step / opposite_step_norm
                    stepsB, stream_status = self._tracker(s, first_step, B)
                    if not (
                        self.return_all
                        or stream_status
                        in (StreamlineStatus.ENDPOINT, StreamlineStatus.OUTSIDEIMAGE)
                    ):
                        continue

                if stepsB == 1:
                    streamline = F[:stepsF].copy()
                else:
                    parts = (B[stepsB - 1 : 0 : -1], F[:stepsF])
                    streamline = np.concatenate(parts, axis=0)

                # move to the next streamline if only the seed position
                # and not return all
                len_sl = len(streamline)
                if (
                    len_sl >= self.min_length
                    and len_sl <= self.max_length
                    or self.return_all
                ):
                    if self.save_seeds:
                        yield streamline, s
                    else:
                        yield streamline


class ParticleFilteringTracking(LocalTracking):
    @warning_for_keywords()
    def __init__(
        self,
        direction_getter,
        stopping_criterion,
        seeds,
        affine,
        step_size,
        *,
        max_cross=None,
        maxlen=500,
        minlen=2,
        pft_back_tracking_dist=2,
        pft_front_tracking_dist=1,
        pft_max_trial=20,
        particle_count=15,
        return_all=True,
        random_seed=None,
        save_seeds=False,
        min_wm_pve_before_stopping=0,
        unidirectional=False,
        randomize_forward_direction=False,
        initial_directions=None,
    ):
        """A streamline generator using the particle filtering tractography
        method.

        See :footcite:p:`Girard2014` for further details about the method.

        Parameters
        ----------
        direction_getter : instance of ProbabilisticDirectionGetter
            Used to get directions for fiber tracking.
        stopping_criterion : instance of AnatomicalStoppingCriterion
            Identifies endpoints and invalid points to inform tracking.
        seeds : array (N, 3)
            Points to seed the tracking. Seed points should be given in point
            space of the track (see ``affine``).
        affine : array (4, 4)
            Coordinate space for the streamline point with respect to voxel
            indices of input data. This affine can contain scaling, rotational,
            and translational components but should not contain any shearing.
            An identity matrix can be used to generate streamlines in "voxel
            coordinates" as long as isotropic voxels were used to acquire the
            data.
        step_size : float
            Step size used for tracking.
        max_cross : int or None, optional
            The maximum number of direction to track from each seed in crossing
            voxels. By default all initial directions are tracked.
        maxlen : int, optional
            Maximum length of generated streamlines. Longer streamlines will be
            discarted if `return_all=False`.
        minlen : int, optional
            Minimum length of generated streamlines. Shorter streamlines will
            be discarted if `return_all=False`.
        pft_back_tracking_dist : float
            Distance in mm to back track before starting the particle filtering
            tractography. The total particle filtering tractography distance is
            equal to back_tracking_dist + front_tracking_dist.
            By default this is set to 2 mm.
        pft_front_tracking_dist : float, optional
            Distance in mm to run the particle filtering tractography after the
            the back track distance. The total particle filtering tractography
            distance is equal to back_tracking_dist + front_tracking_dist. By
            default this is set to 1 mm.
        pft_max_trial : int, optional
            Maximum number of trial for the particle filtering tractography
            (Prevents infinite loops).
        particle_count : int, optional
            Number of particles to use in the particle filter.
        return_all : bool, optional
            If true, return all generated streamlines, otherwise only
            streamlines reaching end points or exiting the image.
        random_seed : int, optional
            The seed for the random seed generator (numpy.random.seed and
            random.seed).
        save_seeds : bool, optional
            If True, return seeds alongside streamlines
        min_wm_pve_before_stopping : int, optional
            Minimum white matter pve (1 - stopping_criterion.include_map -
            stopping_criterion.exclude_map) to reach before allowing the
            tractography to stop.
        unidirectional : bool, optional
            If true, the tracking is performed only in the forward direction.
            The seed position will be the first point of all streamlines.
        randomize_forward_direction : bool, optional
            If true, the forward direction is randomized (multiplied by 1
            or -1). Otherwise, the provided forward direction is used.
        initial_directions: array (N, npeaks, 3), optional
            Initial direction to follow from the ``seed`` position. If
            ``max_cross`` is None, one streamline will be generated per peak
            per voxel. If None, `direction_getter.initial_direction` is used.


        References
        ----------
        .. footbibliography::
        """

        if not isinstance(stopping_criterion, AnatomicalStoppingCriterion):
            raise ValueError("expecting AnatomicalStoppingCriterion")

        self.pft_max_nbr_back_steps = int(np.ceil(pft_back_tracking_dist / step_size))
        self.pft_max_nbr_front_steps = int(np.ceil(pft_front_tracking_dist / step_size))
        pft_max_steps = self.pft_max_nbr_back_steps + self.pft_max_nbr_front_steps

        if (
            self.pft_max_nbr_front_steps < 0
            or self.pft_max_nbr_back_steps < 0
            or pft_max_steps < 1
        ):
            raise ValueError("The number of PFT steps must be greater than 0.")

        if particle_count <= 0:
            raise ValueError("The particle count must be greater than 0.")

        if not 0 <= min_wm_pve_before_stopping <= 1:
            raise ValueError(
                "The min_wm_pve_before_stopping value must be between 0 and 1."
            )

        self.min_wm_pve_before_stopping = min_wm_pve_before_stopping
        self.directions = np.empty((maxlen + 1, 3), dtype=float)
        self.pft_max_trial = pft_max_trial
        self.particle_count = particle_count
        self.particle_paths = np.empty(
            (2, self.particle_count, pft_max_steps + 1, 3), dtype=float
        )
        self.particle_weights = np.empty(self.particle_count, dtype=float)
        self.particle_dirs = np.empty(
            (2, self.particle_count, pft_max_steps + 1, 3), dtype=float
        )
        self.particle_steps = np.empty((2, self.particle_count), dtype=np.intp)
        self.particle_stream_statuses = np.empty(
            (2, self.particle_count), dtype=np.intp
        )
        super(ParticleFilteringTracking, self).__init__(
            direction_getter=direction_getter,
            stopping_criterion=stopping_criterion,
            seeds=seeds,
            affine=affine,
            step_size=step_size,
            max_cross=max_cross,
            maxlen=maxlen,
            minlen=minlen,
            fixedstep=True,
            return_all=return_all,
            random_seed=random_seed,
            save_seeds=save_seeds,
            unidirectional=unidirectional,
            randomize_forward_direction=randomize_forward_direction,
            initial_directions=initial_directions,
        )

    def _tracker(self, seed, first_step, streamline):
        return pft_tracker(
            self.direction_getter,
            self.stopping_criterion,
            seed,
            first_step,
            self._voxel_size,
            streamline,
            self.directions,
            self.step_size,
            self.pft_max_nbr_back_steps,
            self.pft_max_nbr_front_steps,
            self.pft_max_trial,
            self.particle_count,
            self.particle_paths,
            self.particle_dirs,
            self.particle_weights,
            self.particle_steps,
            self.particle_stream_statuses,
            self.min_wm_pve_before_stopping,
        )



"""
=========================================
Tractography Clustering with QuickBundles
=========================================

This example explains how we can use QuickBundles
:footcite:p:`Garyfallidis2012a` to simplify/cluster streamlines.

First import the necessary modules.
"""

import numpy as np

from dipy.data import get_fnames
from dipy.io.pickles import save_pickle
from dipy.io.streamline import load_tractogram
from dipy.segment.clustering import QuickBundles
from dipy.viz import actor, colormap, window

###############################################################################
# For educational purposes we will try to cluster a small streamline bundle
# known from neuroanatomy as the fornix.

fname = get_fnames(name="fornix")

###############################################################################
# Load fornix streamlines.

fornix = load_tractogram(fname, "same", bbox_valid_check=False)
streamlines = fornix.streamlines

###############################################################################
# Perform QuickBundles clustering using the MDF metric and a 10mm distance
# threshold. Keep in mind that since the MDF metric requires streamlines to
# have the same number of points, the clustering algorithm will internally use
# a representation of streamlines that have been automatically
# downsampled/upsampled so they have only 12 points (To set manually the
# number of points, see :ref:`clustering-examples-ResampleFeature`).

qb = QuickBundles(threshold=10.0)
clusters = qb.cluster(streamlines)

###############################################################################
# `clusters` is a `ClusterMap` object which contains attributes that
# provide information about the clustering result.

print("Nb. clusters:", len(clusters))
print("Cluster sizes:", map(len, clusters))
print("Small clusters:", clusters < 10)
print("Streamlines indices of the first cluster:\n", clusters[0].indices)
print("Centroid of the last cluster:\n", clusters[-1].centroid)

###############################################################################
# `clusters` also has attributes such as `centroids` (cluster representatives),
# and methods like `add`, `remove`, and `clear` to modify the clustering
# result.
#
# Let's first show the initial dataset.

# Enables/disables interactive visualization
interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.streamtube(streamlines, colors=window.colors.white))
window.record(scene=scene, out_path="fornix_initial.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Initial Fornix dataset.
#
#
# Show the centroids of the fornix after clustering (with random colors):

colormap = colormap.create_colormap(np.arange(len(clusters)))

scene.clear()
scene.SetBackground(1, 1, 1)
scene.add(actor.streamtube(streamlines, colors=window.colors.white, opacity=0.05))
scene.add(actor.streamtube(clusters.centroids, colors=colormap, linewidth=0.4))
window.record(scene=scene, out_path="fornix_centroids.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Showing the different QuickBundles centroids with random colors.
#
#
# Show the labeled fornix (colors from centroids).

colormap_full = np.ones((len(streamlines), 3))
for cluster, color in zip(clusters, colormap):
    colormap_full[cluster.indices] = color

scene.clear()
scene.SetBackground(1, 1, 1)
scene.add(actor.streamtube(streamlines, colors=colormap_full))
window.record(scene=scene, out_path="fornix_clusters.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Showing the different clusters.
#
#
# It is also possible to save the complete `ClusterMap` object with pickling.

save_pickle("QB.pkl", clusters)

###############################################################################
# Finally, here is a video of QuickBundles applied on a larger dataset.
#
# .. raw:: html
#
#     <iframe width="420" height="315" src="https://www.youtube.com/embed/kstL7KKqu94" frameborder="0" allowfullscreen></iframe>  # noqa: E501
#
#
# References
# ----------
#
# .. footbibliography::
#



"""

.. _reconst-csd:

===================================================================
Reconstruction with Constrained Spherical Deconvolution model (CSD)
===================================================================

This example shows how to use Constrained Spherical Deconvolution (CSD)
introduced by :footcite:p:`Tournier2007`.

This method is mainly useful with datasets with gradient directions acquired on
a spherical grid.

The basic idea with this method is that if we could estimate the response
function of a single fiber then we could deconvolve the measured signal and
obtain the underlying fiber distribution.

In this way, the reconstruction of the fiber orientation distribution function
(fODF) in CSD involves two steps:

    1. Estimation of the fiber response function
    2. Use the response function to reconstruct the fODF

Let's first load the data. We will use a dataset with 10 b0s and 150 non-b0s
with b-value 2000.
"""

import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.direction import peaks_from_model
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst.csdeconv import (
    ConstrainedSphericalDeconvModel,
    auto_response_ssst,
    mask_for_response_ssst,
    recursive_response,
    response_from_mask_ssst,
)
from dipy.reconst.dti import TensorModel, fractional_anisotropy, mean_diffusivity
from dipy.sims.voxel import single_tensor_odf
from dipy.viz import actor, window

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")

data, affine = load_nifti(hardi_fname)

bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# You can verify the b-values of the dataset by looking at the attribute
# ``gtab.bvals``. Now that a dataset with multiple gradient directions is
# loaded, we can proceed with the two steps of CSD.
#
# Step 1. Estimation of the fiber response function
# =================================================
#
# There are many strategies to estimate the fiber response function. Here two
# different strategies are presented.
#
# **Strategy 1 - response function estimates from a local brain region**
# One simple way to estimate the fiber response function is to look for regions
# of the brain where it is known that there are single coherent fiber
# populations. For example, if we use a ROI at the center of the brain, we will
# find single fibers from the corpus callosum. The ``auto_response_ssst``
# function will calculate FA for a cuboid ROI of radii equal to ``roi_radii``
# in the center of the volume and return the response function estimated in
# that region for the voxels with FA higher than 0.7.

response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)

###############################################################################
# Note that the ``auto_response_ssst`` function calls two functions that can be
# used separately. First, the function ``mask_for_response_ssst`` creates a
# mask of voxels within the cuboid ROI that meet the FA threshold constraint.
# This mask can be used to calculate the number of voxels that were kept, or
# to also apply an external mask (a WM mask for example). Second, the function
# ``response_from_mask_ssst`` takes the mask and returns the response function
# calculated within the mask. If no changes are made to the mask between the
# two calls, the resulting responses should be identical.

mask = mask_for_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)
nvoxels = np.sum(mask)
print(nvoxels)

response, ratio = response_from_mask_ssst(gtab, data, mask)

###############################################################################
# The ``response`` tuple contains two elements. The first is an array with
# the eigenvalues of the response function and the second is the average S0 for
# this response.
#
# It is good practice to always validate the result of auto_response_ssst. For
# this purpose we can print the elements of ``response`` and have a look at
# their values.

print(response)

###############################################################################
# The tensor generated from the response must be prolate (two smaller
# eigenvalues should be equal) and look anisotropic with a ratio of second to
# first eigenvalue of about 0.2. Or in other words, the axial diffusivity of
# this tensor should be around 5 times larger than the radial diffusivity.

print(ratio)

###############################################################################
# We can double-check that we have a good response function by visualizing the
# response function's ODF. Here is how you would do that:

# Enables/disables interactive visualization
interactive = False

scene = window.Scene()
evals = response[0]
evecs = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]]).T


response_odf = single_tensor_odf(default_sphere.vertices, evals=evals, evecs=evecs)
# transform our data from 1D to 4D
response_odf = response_odf[None, None, None, :]
response_actor = actor.odf_slicer(
    response_odf, sphere=default_sphere, colormap="plasma"
)
scene.add(response_actor)
print("Saving illustration as csd_response.png")
window.record(scene=scene, out_path="csd_response.png", size=(200, 200))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Estimated response function.

scene.rm(response_actor)

###############################################################################
# **Strategy 2 - data-driven calibration of response function** Depending
# on the dataset, FA threshold may not be the best way to find the best
# possible response function. For one, it depends on the diffusion tensor
# (FA and first eigenvector), which has lower accuracy at high
# b-values. Alternatively, the response function can be calibrated in a
# data-driven manner :footcite:p:`Tax2014`.
#
# First, the data is deconvolved with a 'fat' response function. All voxels
# that are considered to contain only one peak in this deconvolution (as
# determined by the peak threshold which gives an upper limit of the ratio
# of the second peak to the first peak) are maintained, and from these voxels
# a new response function is determined. This process is repeated until
# convergence is reached. Here we calibrate the response function on a small
# part of the data.

###############################################################################
# A WM mask can shorten computation time for the whole dataset. Here it is
# created based on the DTI fit.

tenmodel = TensorModel(gtab)
tenfit = tenmodel.fit(data, mask=data[..., 0] > 200)

FA = fractional_anisotropy(tenfit.evals)
MD = mean_diffusivity(tenfit.evals)
wm_mask = np.logical_or(FA >= 0.4, (np.logical_and(FA >= 0.15, MD >= 0.0011)))

response = recursive_response(
    gtab,
    data,
    mask=wm_mask,
    sh_order_max=8,
    peak_thr=0.01,
    init_fa=0.08,
    init_trace=0.0021,
    iter=8,
    convergence=0.001,
    parallel=True,
    num_processes=2,
)


###############################################################################
# We can check the shape of the signal of the response function, which should
# be like  a pancake:

response_signal = response.on_sphere(default_sphere)
# transform our data from 1D to 4D
response_signal = response_signal[None, None, None, :]
response_actor = actor.odf_slicer(
    response_signal, sphere=default_sphere, colormap="plasma"
)

scene = window.Scene()

scene.add(response_actor)
print("Saving illustration as csd_recursive_response.png")
window.record(scene=scene, out_path="csd_recursive_response.png", size=(200, 200))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Estimated response function using recursive calibration.

scene.rm(response_actor)

###############################################################################
# Step 2. fODF reconstruction
# ===========================
#
# After estimating a response function for one of the strategies shown above,
# we are ready to start the deconvolution process. Let's import the CSD model
# and fit the datasets.

csd_model = ConstrainedSphericalDeconvModel(gtab, response)

###############################################################################
# For illustration purposes we will fit only a small portion of the data.

data_small = data[20:50, 55:85, 38:39]
csd_fit = csd_model.fit(data_small)

###############################################################################
# Show the CSD-based ODFs also known as FODFs (fiber ODFs).

csd_odf = csd_fit.odf(default_sphere)

###############################################################################
# Here we visualize only a 30x30 region.

fodf_spheres = actor.odf_slicer(
    csd_odf, sphere=default_sphere, scale=0.9, norm=False, colormap="plasma"
)

scene.add(fodf_spheres)

print("Saving illustration as csd_odfs.png")
window.record(scene=scene, out_path="csd_odfs.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# CSD ODFs.
#
#
# In DIPY we also provide tools for finding the peak directions (maxima) of the
# ODFs. For this purpose we recommend using ``peaks_from_model``.

csd_peaks = peaks_from_model(
    model=csd_model,
    data=data_small,
    sphere=default_sphere,
    relative_peak_threshold=0.5,
    min_separation_angle=25,
    parallel=True,
    num_processes=2,
)

scene.clear()
fodf_peaks = actor.peak_slicer(csd_peaks.peak_dirs, peaks_values=csd_peaks.peak_values)
scene.add(fodf_peaks)

print("Saving illustration as csd_peaks.png")
window.record(scene=scene, out_path="csd_peaks.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# CSD Peaks.
#
#
# We can finally visualize both the ODFs and peaks in the same space.

fodf_spheres.GetProperty().SetOpacity(0.4)

scene.add(fodf_spheres)

print("Saving illustration as csd_both.png")
window.record(scene=scene, out_path="csd_both.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# CSD Peaks and ODFs.
#
#
# References
# ----------
#
# .. footbibliography::
#



"""This module is dedicated to the handling of tractograms."""

import logging
import os

import numpy as np
import trx.trx_file_memmap as tmm


def concatenate_tractogram(
    tractogram_list,
    *,
    delete_dpv=False,
    delete_dps=False,
    delete_groups=False,
    check_space_attributes=True,
    preallocation=False,
):
    """Concatenate multiple tractograms into one.

    If the data_per_point or data_per_streamline is not the same for all
    tractograms, the data must be deleted first.

    Parameters
    ----------
    tractogram_list : List[StatefulTractogram or TrxFile]
        The stateful tractogram to concatenate
    delete_dpv: bool, optional
        Delete dpv keys that do not exist in all the provided TrxFiles
    delete_dps: bool, optional
        Delete dps keys that do not exist in all the provided TrxFile
    delete_groups: bool, optional
        Delete all the groups that currently exist in the TrxFiles
    check_space_attributes: bool, optional
        Verify that dimensions and size of data are similar between all the
        TrxFiles
    preallocation: bool, optional
        Preallocated TrxFile has already been generated and is the first
        element in trx_list (Note: delete_groups must be set to True as well)

    Returns
    -------
    new_trx: TrxFile
        TrxFile representing the concatenated data

    """
    trx_list = []
    for sft in tractogram_list:
        if not isinstance(sft, tmm.TrxFile):
            sft = tmm.TrxFile.from_sft(sft)
        elif len(sft.groups):
            delete_groups = True
        trx_list.append(sft)

    trx_list = [
        curr_trx for curr_trx in trx_list if curr_trx.header["NB_STREAMLINES"] > 0
    ]

    if not trx_list:
        logging.warning("Inputs of concatenation were empty.")
        return tmm.TrxFile()

    if len(trx_list) == 1:
        if len(tractogram_list) > 1:
            logging.warning("Only 1 valid tractogram returned.")
        return trx_list[0]

    ref_trx = trx_list[0]
    all_dps = []
    all_dpv = []
    for curr_trx in trx_list:
        all_dps.extend(list(curr_trx.data_per_streamline.keys()))
        all_dpv.extend(list(curr_trx.data_per_vertex.keys()))
    all_dps, all_dpv = set(all_dps), set(all_dpv)

    if check_space_attributes:
        for curr_trx in trx_list[1:]:
            if not np.allclose(
                ref_trx.header["VOXEL_TO_RASMM"], curr_trx.header["VOXEL_TO_RASMM"]
            ) or not np.array_equal(
                ref_trx.header["DIMENSIONS"], curr_trx.header["DIMENSIONS"]
            ):
                raise ValueError("Wrong space attributes.")

    if preallocation and not delete_groups:
        raise ValueError("Groups are variables, cannot be handled with preallocation")

    # Verifying the validity of fixed-size arrays, coherence between inputs
    for curr_trx in trx_list:
        for key in all_dpv:
            if (
                key not in ref_trx.data_per_vertex.keys()
                or key not in curr_trx.data_per_vertex.keys()
            ):
                if not delete_dpv:
                    logging.debug(f"{key} dpv key does not exist in all TrxFile.")
                    raise ValueError("TrxFile must be sharing identical dpv keys.")
            elif (
                ref_trx.data_per_vertex[key]._data.dtype
                != curr_trx.data_per_vertex[key]._data.dtype
            ):
                logging.debug(
                    f"{key} dpv key is not declared with the same dtype in all TrxFile."
                )
                raise ValueError("Shared dpv key, has different dtype.")

    for curr_trx in trx_list:
        for key in all_dps:
            if (
                key not in ref_trx.data_per_streamline.keys()
                or key not in curr_trx.data_per_streamline.keys()
            ):
                if not delete_dps:
                    logging.debug(f"{key} dps key does not exist in all TrxFile.")
                    raise ValueError("TrxFile must be sharing identical dps keys.")
            elif (
                ref_trx.data_per_streamline[key].dtype
                != curr_trx.data_per_streamline[key].dtype
            ):
                logging.debug(
                    f"{key} dps key is not declared with the same dtype in all TrxFile."
                )
                raise ValueError("Shared dps key, has different dtype.")

    all_groups_len = {}
    all_groups_dtype = {}
    # Variable-size arrays do not have to exist in all TrxFile
    if not delete_groups:
        for trx_1 in trx_list:
            for group_key in trx_1.groups.keys():
                # Concatenating groups together
                if group_key in all_groups_len:
                    all_groups_len[group_key] += len(trx_1.groups[group_key])
                else:
                    all_groups_len[group_key] = len(trx_1.groups[group_key])
                if (
                    group_key in all_groups_dtype
                    and trx_1.groups[group_key].dtype != all_groups_dtype[group_key]
                ):
                    raise ValueError("Shared group key, has different dtype.")
                else:
                    all_groups_dtype[group_key] = trx_1.groups[group_key].dtype

    # Once the checks are done, actually concatenate
    to_concat_list = trx_list[1:] if preallocation else trx_list
    if not preallocation:
        nb_vertices = 0
        nb_streamlines = 0
        for curr_trx in to_concat_list:
            curr_strs_len, curr_pts_len = curr_trx._get_real_len()
            nb_streamlines += curr_strs_len
            nb_vertices += curr_pts_len

        new_trx = tmm.TrxFile(
            nb_vertices=nb_vertices, nb_streamlines=nb_streamlines, init_as=ref_trx
        )
        if delete_dps:
            new_trx.data_per_streamline = {}
        if delete_dpv:
            new_trx.data_per_vertex = {}
        if delete_groups:
            new_trx.groups = {}

        tmp_dir = new_trx._uncompressed_folder_handle.name

        # When memory is allocated on the spot, groups and data_per_group can
        # be concatenated together
        for group_key in all_groups_len.keys():
            if not os.path.isdir(os.path.join(tmp_dir, "groups/")):
                os.mkdir(os.path.join(tmp_dir, "groups/"))
            dtype = all_groups_dtype[group_key]
            group_filename = os.path.join(tmp_dir, f"groups/{group_key}.{dtype.name}")
            group_len = all_groups_len[group_key]
            new_trx.groups[group_key] = tmm._create_memmap(
                group_filename, mode="w+", shape=(group_len,), dtype=dtype
            )
            if delete_groups:
                continue
            pos = 0
            count = 0
            for curr_trx in trx_list:
                curr_len = len(curr_trx.groups[group_key])
                new_trx.groups[group_key][pos : pos + curr_len] = (
                    curr_trx.groups[group_key] + count
                )
                pos += curr_len
                count += curr_trx.header["NB_STREAMLINES"]

        strs_end, pts_end = 0, 0
    else:
        new_trx = ref_trx
        strs_end, pts_end = new_trx._get_real_len()

    for curr_trx in to_concat_list:
        # Copy the TrxFile fixed-size info (the right chunk)
        strs_end, pts_end = new_trx._copy_fixed_arrays_from(
            curr_trx, strs_start=strs_end, pts_start=pts_end
        )
    return new_trx



import numpy as np
from numpy import zeros
from numpy.testing import assert_equal

from dipy.segment.threshold import upper_bound_by_percent, upper_bound_by_rate


def test_adjustment():
    imga = zeros([128, 128])

    for y in range(128):
        for x in range(128):
            if 10 < y < 115 and 10 < x < 115:
                imga[x, y] = 100
            if 39 < y < 88 and 39 < x < 88:
                imga[x, y] = 150
            if 59 < y < 69 and 59 < x < 69:
                imga[x, y] = 255

    high_1 = upper_bound_by_rate(imga)
    high_2 = upper_bound_by_percent(imga)
    vol1 = np.interp(imga, xp=[imga.min(), high_1], fp=[0, 255])
    vol2 = np.interp(imga, xp=[imga.min(), high_2], fp=[0, 255])
    count2 = (88 - 40) * (88 - 40)
    count1 = (114 - 10) * (114 - 10)

    count1_test = 0
    count2_test = 0

    count2_upper = (88 - 40) * (88 - 40)
    count1_upper = (114 - 10) * (114 - 10)

    count1_upper_test = 0
    count2_upper_test = 0

    value1 = np.unique(vol1)
    value2 = np.unique(vol2)

    for i in range(128):
        for j in range(128):
            if vol1[i][j] > value1[1]:
                count2_test = count2_test + 1
            if vol1[i][j] > 0:
                count1_test = count1_test + 1

    for i in range(128):
        for j in range(128):
            if vol2[i][j] > value2[1]:
                count2_upper_test = count2_upper_test + 1
            if vol2[i][j] > 0:
                count1_upper_test = count1_upper_test + 1

    assert_equal(count2, count2_test)
    assert_equal(count1, count1_test)

    assert_equal(count2_upper, count2_upper_test)
    assert_equal(count1_upper, count1_upper_test)
